import pytest

from fastapi_filters_standard import FilterField, FilterOperator
from fastapi_filters_standard.op import FilterOp

_field_int_a = FilterField(
    name="a",
    type=int,
)
_field_int_b = FilterField(
    name="b",
    type=str | None,
)
_field_arr = FilterField(
    name="arr",
    type=list[int],
)


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        # ==, !=, <, <=, >, >=
        (_field_int_a == 1, FilterOp("a", FilterOperator.eq, 1)),
        (_field_int_a != 1, FilterOp("a", FilterOperator.ne, 1)),
        (_field_int_a < 1, FilterOp("a", FilterOperator.lt, 1)),
        (_field_int_a <= 1, FilterOp("a", FilterOperator.le, 1)),
        (_field_int_a > 1, FilterOp("a", FilterOperator.gt, 1)),
        (_field_int_a >= 1, FilterOp("a", FilterOperator.ge, 1)),
        # eq, ne, lt, le, gt, ge
        (_field_int_a.eq(1), FilterOp("a", FilterOperator.eq, 1)),
        (_field_int_a.ne(1), FilterOp("a", FilterOperator.ne, 1)),
        (_field_int_a.lt(1), FilterOp("a", FilterOperator.lt, 1)),
        (_field_int_a.le(1), FilterOp("a", FilterOperator.le, 1)),
        (_field_int_a.gt(1), FilterOp("a", FilterOperator.gt, 1)),
        (_field_int_a.ge(1), FilterOp("a", FilterOperator.ge, 1)),
        # in, not_in
        (_field_int_a.in_([1, 2]), FilterOp("a", FilterOperator.in_, [1, 2])),
        (_field_int_a >> [1, 2], FilterOp("a", FilterOperator.in_, [1, 2])),
        (_field_int_a.not_in([1, 2]), FilterOp("a", FilterOperator.not_in, [1, 2])),
        # like, not_like, ilike, not_ilike
        (_field_int_b.like("a"), FilterOp("b", FilterOperator.like, "a")),
        (_field_int_b.not_like("a"), FilterOp("b", FilterOperator.not_like, "a")),
        (_field_int_b.ilike("a"), FilterOp("b", FilterOperator.ilike, "a")),
        (_field_int_b.not_ilike("a"), FilterOp("b", FilterOperator.not_ilike, "a")),
        # is_null
        (_field_int_b.is_null(), FilterOp("b", FilterOperator.is_null, True)),
        (_field_int_b.is_null(True), FilterOp("b", FilterOperator.is_null, True)),
        (_field_int_b.is_null(False), FilterOp("b", FilterOperator.is_null, False)),
        # ov, not_ov, contains, not_contains
        (_field_arr.overlaps([1, 2]), FilterOp("arr", FilterOperator.overlap, [1, 2])),
        (
            _field_arr.not_overlaps([1, 2]),
            FilterOp("arr", FilterOperator.not_overlap, [1, 2]),
        ),
        (_field_arr.contains(1), FilterOp("arr", FilterOperator.contains, 1)),
        (_field_arr.not_contains(1), FilterOp("arr", FilterOperator.not_contains, 1)),
        # (eq, 1)
        (_field_int_a(FilterOperator.eq, 1), FilterOp("a", FilterOperator.eq, 1)),
    ],
    ids=[
        "eq-op",
        "ne-op",
        "lt-op",
        "le-op",
        "gt-op",
        "ge-op",
        "eq-func",
        "ne-func",
        "lt-func",
        "le-func",
        "gt-func",
        "ge-func",
        "in-func",
        "not-in-func",
        "in-op",
        "like-func",
        "not-like-func",
        "ilike-func",
        "not-ilike-func",
        "is-null-func",
        "is-null-func-true",
        "is-null-func-false",
        "ov-func",
        "not-ov-func",
        "contains-func",
        "not-contains-func",
        "call",
    ],
)
def test_op_constructor(expr, expected):
    assert expr == expected


def test_missed_fields():
    _field_no_name = FilterField(type=int)

    with pytest.raises(
        AssertionError,
        match=r"^FilterField has no name$",
    ):
        _ = _field_no_name == 1

    _field_no_type = FilterField(name="a")

    with pytest.raises(
        AssertionError,
        match=r"^FilterField has no operators$",
    ):
        _ = _field_no_type == 1



from datetime import datetime
from typing import Any
from uuid import UUID, uuid4

from faker import Faker
from fastapi import Depends, FastAPI
from pydantic import BaseModel
from sqlalchemy import Integer, func, select, text
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, mapped_column

from fastapi_filters_standard import (
    FilterField,
    FilterSet,
    SortingValues,
    create_filters_from_set,
    create_sorting,
)
from fastapi_filters_standard.ext.sqlalchemy import apply_filters_and_sorting

faker = Faker()
engine = create_async_engine(
    "postgresql+asyncpg://postgres:postgres@localhost:5432/postgres",
)


class Base(MappedAsDataclass, DeclarativeBase):
    pass


class User(Base):
    __tablename__ = "users"

    id: Mapped[UUID] = mapped_column(primary_key=True)

    first_name: Mapped[str] = mapped_column()
    last_name: Mapped[str] = mapped_column()
    email: Mapped[str] = mapped_column()
    age: Mapped[int] = mapped_column()
    marks: Mapped[list[int]] = mapped_column(ARRAY(Integer))
    is_active: Mapped[bool] = mapped_column()
    created_at: Mapped[datetime] = mapped_column()


class UserOut(BaseModel):
    id: UUID
    first_name: str
    last_name: str
    email: str
    age: int
    marks: list[int]
    is_active: bool
    created_at: datetime

    model_config = {
        "from_attributes": True,
    }


app = FastAPI()


@app.on_event("startup")
async def on_startup() -> None:
    async with engine.begin() as conn:
        await conn.execute(text("drop schema public cascade"))
        await conn.execute(text("create schema public"))

        await conn.run_sync(Base.metadata.create_all)

    async with AsyncSession(engine) as session, session.begin():
        session.add_all(
            [
                User(
                    id=uuid4(),
                    first_name=faker.first_name(),
                    last_name=faker.last_name(),
                    email=faker.email(),
                    marks=[
                        faker.pyint(min_value=1, max_value=5) for _ in range(faker.pyint(min_value=1, max_value=10))
                    ],
                    age=faker.pyint(min_value=1, max_value=100),
                    is_active=faker.pybool(),
                    created_at=faker.date_time_this_year(),
                )
                for _ in range(100)
            ],
        )


class UserFiltersSet(FilterSet):
    first_name: FilterField[str]
    last_name: FilterField[str]
    email: FilterField[str]
    age: FilterField[int]
    marks: FilterField[list[int]]
    is_active: FilterField[bool]
    created_at: FilterField[datetime]


@app.get("/users", response_model=list[UserOut])
async def get_users(
    filters: UserFiltersSet = Depends(create_filters_from_set(UserFiltersSet)),
    sorting: SortingValues = Depends(create_sorting("age", "created_at")),
) -> Any:
    stmt = select(User)

    stmt = apply_filters_and_sorting(
        stmt,
        filters,
        sorting,
        additional={
            "first_name": func.lower(User.first_name),
            "last_name": func.lower(User.last_name),
        },
    )

    async with AsyncSession(engine) as session:
        return await session.scalars(stmt)


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app)



import re
from pathlib import Path
from subprocess import run
from sys import executable
from textwrap import dedent
from typing import Any

from pydantic import TypeAdapter

LINE_START_REGEX = re.compile(r"(?m)^.*?\.py|", re.MULTILINE)


def parse_obj_as(tp: Any, obj: Any) -> Any:
    return TypeAdapter(tp).validate_python(obj)


def _normalize_str(s: str) -> str:
    return dedent(s).strip()
    # return LINE_START_REGEX.sub("", s)


def _run_mypy_process(path: Path, params: str = "") -> str:
    cmd = f"{executable} -m mypy.__main__ --strict {params} {path}"

    process = run(cmd, shell=True, capture_output=True, timeout=30, check=False)
    return (process.stdout or process.stderr).decode()


def run_mypy(path: Path, expected: str) -> tuple[bool, str]:
    actual = _run_mypy_process(path)

    actual = _normalize_str(actual)
    expected = _normalize_str(expected)

    if actual == expected:
        return True, ""

    out = "\n".join(
        [
            f"Path: {path.name}",
            "Actual:",
            f"{actual}",
            "Expected:",
            f"{expected}",
        ],
    )

    return False, out



from __future__ import annotations

from collections.abc import Callable, Container, Iterator
from datetime import date, datetime, timedelta
from enum import Enum
from typing import TYPE_CHECKING

from .config import ConfigVar
from .utils import (
    is_optional,
    is_seq,
    lenient_issubclass,
    unwrap_annotated,
    unwrap_optional_type,
    unwrap_type,
)

if TYPE_CHECKING:
    from .types import AbstractFilterOperator


class FilterOperator(str, Enum):
    eq = "eq"
    ne = "ne"
    gt = "gt"
    ge = "ge"
    lt = "lt"
    le = "le"
    like = "like"
    not_like = "not_like"
    ilike = "ilike"
    not_ilike = "not_ilike"
    in_ = "in"
    not_in = "not_in"
    is_null = "is_null"
    overlap = "overlap"
    not_overlap = "not_overlap"
    contains = "contains"
    not_contains = "not_contains"

    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.name}"


DEFAULT_OPERATORS = [
    FilterOperator.eq,
    FilterOperator.ne,
    FilterOperator.in_,
    FilterOperator.not_in,
]

NUM_OPERATORS = [
    FilterOperator.gt,
    FilterOperator.ge,
    FilterOperator.lt,
    FilterOperator.le,
]

BOOL_OPERATORS = [
    FilterOperator.eq,
    FilterOperator.ne,
]

STR_OPERATORS = [
    FilterOperator.like,
    FilterOperator.ilike,
    FilterOperator.not_like,
    FilterOperator.not_ilike,
]

SEQ_OPERATORS = [
    FilterOperator.overlap,
    FilterOperator.not_overlap,
    FilterOperator.contains,
    FilterOperator.not_contains,
]


def default_filter_operators_generator(t: type) -> Iterator[AbstractFilterOperator]:
    t = unwrap_annotated(t)

    if is_optional(t):
        t = unwrap_optional_type(t)
        yield FilterOperator.is_null

    if is_seq(t):
        yield from SEQ_OPERATORS
        return

    tp = unwrap_type(t)

    if lenient_issubclass(tp, bool):
        yield from BOOL_OPERATORS
        return

    yield from DEFAULT_OPERATORS

    if lenient_issubclass(tp, str):
        yield from STR_OPERATORS

    if lenient_issubclass(tp, (int, float, date, datetime, timedelta)):
        yield from NUM_OPERATORS


disabled_filters_config: ConfigVar[Container[AbstractFilterOperator]] = ConfigVar(
    "disabled_filters",
    default=(),
)
filter_operators_generator_config: ConfigVar[Callable[[type], Iterator[AbstractFilterOperator]]] = ConfigVar(
    "filter_operators_generator",
    default=default_filter_operators_generator,
)


def get_filter_operators(t: type) -> Iterator[AbstractFilterOperator]:
    disabled = disabled_filters_config.get()
    operator_generator = filter_operators_generator_config.get()

    for op in operator_generator(t):
        if op not in disabled:
            yield op


__all__ = [
    "DEFAULT_OPERATORS",
    "NUM_OPERATORS",
    "SEQ_OPERATORS",
    "STR_OPERATORS",
    "FilterOperator",
    "default_filter_operators_generator",
    "disabled_filters_config",
    "filter_operators_generator_config",
    "get_filter_operators",
]



from dataclasses import dataclass, fields
from typing import Any

import pytest
from fastapi import Depends, status

from fastapi_filters_standard import (
    FilterField,
    FilterOperator,
    FilterSet,
    create_filters,
    create_filters_from_set,
)
from fastapi_filters_standard.utils import unwrap_annotated


@dataclass
class _CmpFilterField:  # noqa: PLW1641
    val: FilterField[Any]

    def __eq__(self, other: FilterField[Any]) -> bool:
        return fields(self.val) == fields(other)


def test_filter_set_decl():
    class _FilterSet(FilterSet):
        a: FilterField[int]
        b: FilterField[str] = FilterField()
        c: FilterField[bool] = FilterField(
            default_op=FilterOperator.ne,
            operators=[FilterOperator.eq, FilterOperator.ne],
        )

    _fields = {
        "a": FilterField(int, name="a"),
        "b": FilterField(str, name="b"),
        "c": FilterField(
            bool,
            name="c",
            default_op=FilterOperator.ne,
            operators=[FilterOperator.eq, FilterOperator.ne],
        ),
    }

    assert _FilterSet.__filters__ == {k: _CmpFilterField(v) for k, v in _fields.items()}


def test_filter_set_decl_mult_inheritance():
    class _Base1(FilterSet):
        a: FilterField[int]

    class _Base2(FilterSet):
        b: FilterField[str]

    class _FilterSet(_Base1, _Base2):
        c: FilterField[bool]

    _fields = {
        "a": FilterField(int, name="a"),
        "b": FilterField(str, name="b"),
        "c": FilterField(bool, name="c"),
    }

    assert _FilterSet.__filters__ == {k: _CmpFilterField(v) for k, v in _fields.items()}


def test_filter_field_filter_values():
    class _FilterSet(FilterSet):
        a: FilterField[int]
        b: FilterField[str]

    _filter_set = _FilterSet(a={FilterOperator.eq: 1, FilterOperator.ne: 2}, b=None)

    assert _filter_set.filter_values == {
        "a": {FilterOperator.eq: 1, FilterOperator.ne: 2},
    }


def test_missed_field():
    class _FilterSet(FilterSet):
        a: FilterField[int]
        b: FilterField[str]

    _filter_set = _FilterSet(a={FilterOperator.eq: 1})

    assert _filter_set.a == {FilterOperator.eq: 1}
    assert _filter_set.b == {}


def test_from_ops():
    class _FilterSet(FilterSet):
        a: FilterField[int]

    _filter_set = _FilterSet.from_ops(
        _FilterSet.a == 1,
        _FilterSet.a != 2,
    )

    assert _filter_set == _FilterSet(a={FilterOperator.eq: 1, FilterOperator.ne: 2})


def test_from_ops_duplicates():
    class _FilterSet(FilterSet):
        a: FilterField[int]

    with pytest.raises(
        ValueError,
        match=r"^Duplicate operator eq for a$",
    ):
        _FilterSet.from_ops(
            _FilterSet.a == 1,
            _FilterSet.a == 2,
        )


def test_from_resolver():
    _filter_set = FilterSet.create_from_resolver(create_filters(a=int, b=str))

    _fields = {
        "a": FilterField(int, name="a"),
        "b": FilterField(str, name="b"),
    }

    assert issubclass(_filter_set, FilterSet)
    assert _filter_set.__filters__ == {k: _CmpFilterField(v) for k, v in _fields.items()}


@pytest.mark.asyncio
async def test_create_filters_from_set(app, client):
    class _FilterSet(FilterSet):
        a: FilterField[int]
        b: FilterField[str]

    @app.get("/test")
    async def route(
        _filters: _FilterSet = Depends(create_filters_from_set(_FilterSet)),
    ):
        assert isinstance(_filters, _FilterSet)
        assert _filters == _FilterSet(
            a={FilterOperator.eq: 1, FilterOperator.ne: 2},
            b={FilterOperator.eq: "a", FilterOperator.ne: "b"},
        )

        return {}

    response = await client.get(
        "/test",
        params={
            "a": "1",
            "a__ne": "2",
            "b": "a",
            "b__ne": "b",
        },
    )
    assert response.status_code == status.HTTP_200_OK


@pytest.mark.asyncio
async def test_filterset_in_decl(app, client):
    class _FilterSet(FilterSet):
        a: FilterField[int]
        b: FilterField[str]

    @app.get("/test")
    async def route(
        _filters: _FilterSet = Depends(),
    ):
        assert isinstance(_filters, _FilterSet)
        assert _filters == _FilterSet(
            a={FilterOperator.eq: 1, FilterOperator.ne: 2},
            b={FilterOperator.eq: "a", FilterOperator.ne: "b"},
        )

        return {}

    response = await client.get(
        "/test",
        params={
            "a": "1",
            "a__ne": "2",
            "b": "a",
            "b__ne": "b",
        },
    )
    assert response.status_code == status.HTTP_200_OK


def test_subset():
    class _FilterSet(FilterSet):
        a: FilterField[int]
        b: FilterField[str]
        c: FilterField[bool]

    _filters = _FilterSet(
        a={FilterOperator.eq: 1, FilterOperator.ne: 2},
        c={FilterOperator.eq: True, FilterOperator.ne: False},
    )

    assert _filters.subset("a").filter_values == {
        "a": {FilterOperator.eq: 1, FilterOperator.ne: 2},
    }
    assert _filters.subset("b").filter_values == {}
    assert _filters.subset(_FilterSet.a, _FilterSet.b).filter_values == {
        "a": {FilterOperator.eq: 1, FilterOperator.ne: 2},
    }


def test_extract():
    class _BaseFilterSet(FilterSet):
        a: FilterField[int]
        b: FilterField[str]

    class _FilterSet(_BaseFilterSet):
        c: FilterField[bool]

    _filters = _FilterSet(
        a={FilterOperator.eq: 1, FilterOperator.ne: 2},
        b={FilterOperator.eq: "a", FilterOperator.ne: "b"},
        c={FilterOperator.eq: True, FilterOperator.ne: False},
    )

    extracted = _filters.extract(_FilterSet.a)

    assert extracted.filter_values == {
        "a": {FilterOperator.eq: 1, FilterOperator.ne: 2},
    }
    assert _filters.filter_values == {
        "b": {FilterOperator.eq: "a", FilterOperator.ne: "b"},
        "c": {FilterOperator.eq: True, FilterOperator.ne: False},
    }

    assert _filters.extract("d").filter_values == {}

    _filters = _FilterSet(
        a={FilterOperator.eq: 1, FilterOperator.ne: 2},
        b={FilterOperator.eq: "a", FilterOperator.ne: "b"},
        c={FilterOperator.eq: True, FilterOperator.ne: False},
    )

    assert _filters.extract(_BaseFilterSet).filter_values == {
        "a": {FilterOperator.eq: 1, FilterOperator.ne: 2},
        "b": {FilterOperator.eq: "a", FilterOperator.ne: "b"},
    }


def test_bool():
    class _FilterSet(FilterSet):
        a: FilterField[int]
        b: FilterField[str]
        c: FilterField[bool]

    assert not _FilterSet.create()
    assert _FilterSet.create(a={FilterOperator.eq: 1})
    assert _FilterSet.create(b={FilterOperator.eq: "a"})
    assert _FilterSet.create(c={FilterOperator.eq: True})


def test_op_types():
    class _FilterSet(FilterSet):
        a: FilterField[int] = FilterField(
            op_types={
                FilterOperator.eq: float,
                FilterOperator.ne: float,
            },
        )

    resolver = create_filters_from_set(_FilterSet)
    attrs = {f.name: unwrap_annotated(f.type) for f in fields(resolver.__model__)}

    assert attrs == {
        "a": float,
        "a__eq": float,
        "a__ne": float,
        "a__in": list[int],
        "a__not_in": list[int],
        "a__gt": int,
        "a__gte": int,
        "a__lt": int,
        "a__lte": int,
    }



from collections import defaultdict
from collections.abc import Container, Iterator
from dataclasses import asdict, make_dataclass
from typing import (
    Annotated,
    Any,
    cast,
    get_args,
    get_origin,
)

from fastapi import Depends, Query
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from .config import ConfigVar
from .fields import FilterField
from .operators import FilterOperator
from .schemas import CSVList
from .types import (
    AbstractFilterOperator,
    FilterAliasGenerator,
    FilterFieldDef,
    FilterPlace,
    FiltersResolver,
    FilterValues,
)
from .utils import async_safe, fields_include_exclude, is_seq, unwrap_annotated, unwrap_seq_type

# Django-style lookup expressions mapping
LOOKUP_EXPRESSIONS = {
    FilterOperator.eq: "",  # exact - no suffix
    FilterOperator.ne: "__ne",  # not equal
    FilterOperator.gt: "__gt",
    FilterOperator.ge: "__gte",
    FilterOperator.lt: "__lt",
    FilterOperator.le: "__lte",
    FilterOperator.like: "__contains",  # like -> contains
    FilterOperator.not_like: "__not_like",
    FilterOperator.ilike: "__icontains",  # ilike -> icontains
    FilterOperator.not_ilike: "__not_ilike",
    FilterOperator.in_: "__in",
    FilterOperator.not_in: "__not_in",
    FilterOperator.is_null: "__isnull",
    FilterOperator.overlap: "__overlap",
    FilterOperator.not_overlap: "__not_overlap",
    FilterOperator.contains: "__contains",
    FilterOperator.not_contains: "__not_contains",
}


def default_alias_generator(
    name: str,
    op: AbstractFilterOperator,
    alias: str | None = None,
) -> str:
    name = alias or name
    lookup_expr = LOOKUP_EXPRESSIONS.get(op, f"__{op.name.rstrip('_')}")

    # If lookup_expr is empty (exact match), return just the name
    if not lookup_expr:
        return name

    return f"{name}{lookup_expr}"


alias_generator_config: ConfigVar[FilterAliasGenerator] = ConfigVar(
    "alias_generator",
    default=default_alias_generator,
)


def _is_csv_list_type(tp: Any) -> bool:
    """Check if type is CSVList (Annotated[list[T], ...])."""
    # CSVList is Annotated[list[T], ...], so we need to unwrap Annotated
    origin = get_origin(tp)
    
    # If it's Annotated, get the first argument (the actual type)
    if origin is Annotated:
        args = get_args(tp)
        if args:
            inner_type = args[0]
            # Check if inner type is list
            inner_origin = get_origin(inner_type)
            if inner_origin is list:
                return True
            if inner_type is list:
                return True
    
    # Check if it's directly a list type
    if origin is list:
        return True
    if tp is list:
        return True
    
    return False


def _create_query_param(alias: str | None, tp: Any, op: AbstractFilterOperator, field: FilterField[Any], original_tp: type[Any], in_: FilterPlace) -> Any:
    """Create Query parameter with explode=False for CSVList types."""
    # Check if this operation needs CSVList:
    # 1. in_ and not_in operations always use CSVList
    # 2. Sequence types (list, tuple, etc.) use CSVList
    # 3. If type is already CSVList
    needs_csv = (
        op in {FilterOperator.in_, FilterOperator.not_in}
        or is_seq(original_tp)
        or _is_csv_list_type(tp)
    )
    
    if needs_csv:
        if alias:
            return in_(alias=alias, explode=False)
        return in_(explode=False)
    if alias:
        return in_(alias=alias)
    return in_


def adapt_type(
    field: FilterField[Any],
    tp: type[Any],
    op: AbstractFilterOperator,
) -> Any:
    if field.op_types and op in field.op_types:
        return field.op_types[op]

    if is_seq(tp):
        return CSVList[unwrap_seq_type(tp)]  # type: ignore[misc]

    if op in {
        FilterOperator.like,
        FilterOperator.ilike,
        FilterOperator.not_like,
        FilterOperator.not_ilike,
    }:
        return str

    if op == FilterOperator.is_null:
        return bool

    if op in {FilterOperator.in_, FilterOperator.not_in}:
        return CSVList[tp]  # type: ignore[valid-type]

    return tp


def _get_field_name(name: str, op: AbstractFilterOperator) -> str:
    """Get field name for dataclass field, ensuring uniqueness."""
    lookup_expr = LOOKUP_EXPRESSIONS.get(op, f"__{op.name.rstrip('_')}")

    # If lookup_expr is empty (exact match), use __{op.name} for uniqueness
    if not lookup_expr:
        return f"{name}__{op.name}"

    # Remove leading __ from lookup_expr and add it to name
    if lookup_expr.startswith("__"):
        return f"{name}{lookup_expr}"

    return f"{name}__{lookup_expr}"


def field_filter_to_raw_fields(
    name: str,
    field: FilterField[Any],
    alias_generator: FilterAliasGenerator | None = None,
) -> Iterator[tuple[str, Any, AbstractFilterOperator, str | None]]:
    if alias_generator is None:
        alias_generator = alias_generator_config.get()

    yield name, field.type, cast(AbstractFilterOperator, field.default_op), field.alias

    for op in field.operators or ():
        yield (
            _get_field_name(name, op),
            field.type,
            op,
            alias_generator(
                name,
                op,
                field.alias,
            ),
        )


def create_filters_from_model(
    model: type[BaseModel],
    *,
    in_: FilterPlace | None = None,
    alias_generator: FilterAliasGenerator | None = None,
    include: Container[str] | None = None,
    exclude: Container[str] | None = None,
    **overrides: FilterFieldDef,
) -> FiltersResolver:
    checker = fields_include_exclude(model.model_fields, include, exclude)

    def _get_type(f: FieldInfo) -> Any:
        if f.metadata:
            return Annotated[(f.annotation, *f.metadata)]

        return f.annotation

    return create_filters(
        in_=in_,
        alias_generator=alias_generator,
        **{
            **{name: _get_type(field) for name, field in model.model_fields.items() if checker(name)},
            **(overrides or {}),
        },
    )


def create_filters(
    *,
    in_: FilterPlace | None = None,
    alias_generator: FilterAliasGenerator | None = None,
    **kwargs: FilterFieldDef,
) -> FiltersResolver:
    if in_ is None:
        in_ = Query

    fields: dict[str, FilterField[Any]] = {
        name: f_def if isinstance(f_def, FilterField) else FilterField(f_def) for name, f_def in kwargs.items()
    }

    fields_defs = [
        (name, fname, field, tp, alias, op)
        for name, field in fields.items()
        for fname, tp, op, alias in field_filter_to_raw_fields(
            name,
            field,
            alias_generator,
        )
    ]

    defs = {fname: (name, op) for name, fname, *_, op in fields_defs}

    filter_model = make_dataclass(
        "Filters",
        [
            (
                fname,
                Annotated[
                    adapt_type(field, tp, op),
                    _create_query_param(alias, adapt_type(field, tp, op), op, field, tp, in_),
                ],
                None,
            )
            for _, fname, field, tp, alias, op in fields_defs
        ],
    )

    async def _get_filters(f: Any = Depends(async_safe(filter_model))) -> FilterValues:
        values: FilterValues = defaultdict(dict)

        for key, value in asdict(f).items():
            if value is not None:
                name, op = defs[key]
                values[name][op] = value

        return {**values}

    _get_filters.__model__ = filter_model  # type: ignore[attr-defined]
    _get_filters.__defs__ = defs  # type: ignore[attr-defined]
    _get_filters.__filters__ = fields  # type: ignore[attr-defined]

    return cast(FiltersResolver, _get_filters)


__all__ = [
    "alias_generator_config",
    "create_filters",
    "create_filters_from_model",
]



import pytest
from dirty_equals import IsPartialDict
from fastapi import Query, status
from pydantic import ValidationError

from fastapi_filters_standard.schemas import CSVList

from .utils import parse_obj_as


def test_csv_list():
    assert parse_obj_as(CSVList[int], "1") == [1]
    assert parse_obj_as(CSVList[int], "1,2") == [1, 2]
    assert parse_obj_as(CSVList[int], "1,2,3") == [1, 2, 3]


def test_csv_list_with_multiple_values():
    """Test that multiple values (from repeated query params) are handled correctly."""
    # Simulate FastAPI receiving multiple values as a list
    assert parse_obj_as(CSVList[int], ["1", "2", "3"]) == [1, 2, 3]
    assert parse_obj_as(CSVList[str], ["a", "b", "c"]) == ["a", "b", "c"]


def test_csv_list_errors():
    with pytest.raises(ValidationError) as exc:
        parse_obj_as(CSVList[int], "abc")

    assert exc.value.errors() == [
        IsPartialDict(
            {
                "input": "abc",
                "loc": (0,),
                "msg": "Input should be a valid integer, unable to parse string as an integer",
                "type": "int_parsing",
            },
        ),
    ]


@pytest.mark.asyncio
async def test_csv_list_as_query_param(app, client):
    @app.get("/")
    def index(q: CSVList[int] = Query(...)):
        return q

    res = await client.get("/", params={"q": "1"})

    assert res.status_code == status.HTTP_200_OK
    assert res.json() == [1]

    res = await client.get("/", params={"q": "1,abc"})

    assert res.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY
    assert res.json() == {
        "detail": [
            IsPartialDict(
                {
                    "input": "1,abc",
                    "loc": ["query", "q", 0],
                    "msg": "Input should be a valid integer, unable to parse string as an integer",
                    "type": "int_parsing",
                },
            ),
        ],
    }



from contextlib import AsyncExitStack

import pytest
from asgi_lifespan import LifespanManager
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from pytest_asyncio import fixture as async_fixture

import fastapi_filters_standard.configs  # noqa: F401


@pytest.fixture
def app():
    return FastAPI()


@async_fixture
async def client(app):
    async with AsyncExitStack() as astack:
        await astack.enter_async_context(LifespanManager(app))
        c = await astack.enter_async_context(
            AsyncClient(
                transport=ASGITransport(app),
                base_url="http://testserver",
            ),
        )

        yield c



from typing import Annotated

import pytest
from fastapi import Depends, status
from pydantic import BaseModel, BeforeValidator

from fastapi_filters_standard import (
    FilterField,
    FilterOperator,
    FilterValues,
    create_filters,
    create_filters_from_model,
)


@pytest.mark.asyncio
async def test_filters_as_dep(app, client):
    @app.get("/")
    async def route(
        filters: FilterValues = Depends(
            create_filters(
                a=int,
                b=bool,
                c=list[str],
                d=FilterField(
                    bytes,
                    default_op=FilterOperator.eq,
                    operators=[FilterOperator.eq, FilterOperator.ne],
                    alias="d_alias",
                ),
            ),
        ),
    ) -> FilterValues:
        return filters

    res = await client.get(
        "/",
        params={"a": 1, "b": "true", "c": "a,b,c", "d_alias": "123"},
    )

    assert res.status_code == status.HTTP_200_OK
    assert res.json() == {
        "a": {"eq": 1},
        "b": {"eq": True},
        "c": {"overlap": ["a", "b", "c"]},
        "d": {"eq": "123"},
    }

    res = await client.get("/", params={"a": 1, "a__gt": 2, "a__lt": 3})

    assert res.status_code == status.HTTP_200_OK
    assert res.json() == {
        "a": {"eq": 1, "gt": 2, "lt": 3},
    }


@pytest.mark.asyncio
async def test_filters_in_operator_with_comma_separated_values(app, client):
    @app.get("/")
    async def route(
        filters: FilterValues = Depends(
            create_filters(
                main_industry=int,
            ),
        ),
    ) -> FilterValues:
        return filters

    # Correct: comma-separated values in single parameter
    res = await client.get("/", params={"main_industry__in": "1,2,3"})

    assert res.status_code == status.HTTP_200_OK
    assert res.json() == {
        "main_industry": {"in": [1, 2, 3]},
    }

    # Test that repeated parameters are handled correctly
    # FastAPI may receive multiple values as a list, validator should handle it
    res = await client.get("/", params=[("main_industry__in", "1"), ("main_industry__in", "2"), ("main_industry__in", "3")])

    # The validator should join multiple values and split them
    assert res.status_code == status.HTTP_200_OK
    assert res.json() == {
        "main_industry": {"in": [1, 2, 3]},
    }


@pytest.mark.asyncio
async def test_filters(app):
    resolver = create_filters(
        a=FilterField(
            int,
            default_op=FilterOperator.ne,
            operators=[FilterOperator.eq, FilterOperator.ne],
        ),
        b=float,
        c=str | None,
    )

    assert resolver.__defs__ == {
        "a": ("a", FilterOperator.ne),
        "a__eq": ("a", FilterOperator.eq),
        "a__ne": ("a", FilterOperator.ne),
        "b": ("b", FilterOperator.eq),
        "b__eq": ("b", FilterOperator.eq),
        "b__gte": ("b", FilterOperator.ge),
        "b__gt": ("b", FilterOperator.gt),
        "b__in": ("b", FilterOperator.in_),
        "b__lte": ("b", FilterOperator.le),
        "b__lt": ("b", FilterOperator.lt),
        "b__ne": ("b", FilterOperator.ne),
        "b__not_in": ("b", FilterOperator.not_in),
        "c": ("c", FilterOperator.eq),
        "c__eq": ("c", FilterOperator.eq),
        "c__in": ("c", FilterOperator.in_),
        "c__isnull": ("c", FilterOperator.is_null),
        "c__ne": ("c", FilterOperator.ne),
        "c__not_in": ("c", FilterOperator.not_in),
        "c__contains": ("c", FilterOperator.like),
        "c__not_like": ("c", FilterOperator.not_like),
        "c__icontains": ("c", FilterOperator.ilike),
        "c__not_ilike": ("c", FilterOperator.not_ilike),
    }


@pytest.mark.asyncio
async def test_filters_from_model(app):
    class UserModel(BaseModel):
        id: int
        name: Annotated[str, BeforeValidator(lambda v: v)]
        is_active: bool

    assert create_filters_from_model(UserModel).__defs__ == {
        "id": ("id", FilterOperator.eq),
        "id__eq": ("id", FilterOperator.eq),
        "id__gte": ("id", FilterOperator.ge),
        "id__gt": ("id", FilterOperator.gt),
        "id__in": ("id", FilterOperator.in_),
        "id__lte": ("id", FilterOperator.le),
        "id__lt": ("id", FilterOperator.lt),
        "id__ne": ("id", FilterOperator.ne),
        "id__not_in": ("id", FilterOperator.not_in),
        "is_active": ("is_active", FilterOperator.eq),
        "is_active__eq": ("is_active", FilterOperator.eq),
        "is_active__ne": ("is_active", FilterOperator.ne),
        "name": ("name", FilterOperator.eq),
        "name__eq": ("name", FilterOperator.eq),
        "name__in": ("name", FilterOperator.in_),
        "name__ne": ("name", FilterOperator.ne),
        "name__not_in": ("name", FilterOperator.not_in),
        "name__contains": ("name", FilterOperator.like),
        "name__not_like": ("name", FilterOperator.not_like),
        "name__icontains": ("name", FilterOperator.ilike),
        "name__not_ilike": ("name", FilterOperator.not_ilike),
    }
    assert create_filters_from_model(UserModel, include={"name"}).__defs__ == {
        "name": ("name", FilterOperator.eq),
        "name__eq": ("name", FilterOperator.eq),
        "name__in": ("name", FilterOperator.in_),
        "name__ne": ("name", FilterOperator.ne),
        "name__not_in": ("name", FilterOperator.not_in),
        "name__contains": ("name", FilterOperator.like),
        "name__not_like": ("name", FilterOperator.not_like),
        "name__icontains": ("name", FilterOperator.ilike),
        "name__not_ilike": ("name", FilterOperator.not_ilike),
    }
    assert create_filters_from_model(UserModel, exclude={"id"}).__defs__ == {
        "is_active": ("is_active", FilterOperator.eq),
        "is_active__eq": ("is_active", FilterOperator.eq),
        "is_active__ne": ("is_active", FilterOperator.ne),
        "name": ("name", FilterOperator.eq),
        "name__eq": ("name", FilterOperator.eq),
        "name__in": ("name", FilterOperator.in_),
        "name__ne": ("name", FilterOperator.ne),
        "name__not_in": ("name", FilterOperator.not_in),
        "name__contains": ("name", FilterOperator.like),
        "name__not_like": ("name", FilterOperator.not_like),
        "name__icontains": ("name", FilterOperator.ilike),
        "name__not_ilike": ("name", FilterOperator.not_ilike),
    }



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from datetime import datetime
from typing import Any

from beanie import Document, init_beanie
from faker import Faker
from fastapi import Depends, FastAPI
from motor.motor_asyncio import AsyncIOMotorClient

from fastapi_filters_standard import (
    FilterField,
    FilterSet,
    SortingValues,
    create_filters_from_set,
    create_sorting,
)
from fastapi_filters_standard.ext.beanie import apply_filters_and_sorting

faker = Faker()


class User(Document):
    first_name: str
    last_name: str
    email: str
    age: int
    marks: list[int]
    is_active: bool
    created_at: datetime

    class Settings:
        name = "users"


@asynccontextmanager
async def lifespan(_: Any) -> AsyncIterator[None]:
    client = AsyncIOMotorClient()
    await init_beanie(database=client.test, document_models=[User])

    await User.delete_all()

    users = [
        User(
            first_name=faker.first_name(),
            last_name=faker.last_name(),
            email=faker.email(),
            marks=[faker.pyint(min_value=1, max_value=5) for _ in range(faker.pyint(min_value=1, max_value=10))],
            age=faker.pyint(min_value=1, max_value=100),
            is_active=faker.pybool(),
            created_at=faker.date_time_this_year(),
        )
        for _ in range(100)
    ]

    for user in users:
        await user.insert()

    try:
        yield
    finally:
        client.close()


app = FastAPI(lifespan=lifespan)


class UserFiltersSet(FilterSet):
    first_name: FilterField[str]
    last_name: FilterField[str]
    email: FilterField[str]
    age: FilterField[int]
    marks: FilterField[list[int]]
    is_active: FilterField[bool]
    created_at: FilterField[datetime]


@app.get("/users")
async def get_users(
    filters: UserFiltersSet = Depends(create_filters_from_set(UserFiltersSet)),
    sorting: SortingValues = Depends(create_sorting("age", "created_at")),
) -> list[User]:
    stmt = User.find()

    stmt = apply_filters_and_sorting(
        stmt,
        filters,
        sorting,
    )

    return await stmt.to_list()


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app)



import pytest
from fastapi import Depends, status

from fastapi_filters_standard.config import ConfigVar


def test_config_var():
    var = ConfigVar("test", default=1)

    assert var.get() == 1

    with var.set(2):
        assert var.get() == 2

    assert var.get() == 1

    with var.set(3) as reset_ctx:
        assert var.get() == 3
        reset_ctx.reset()
        assert var.get() == 1


def test_config_var_double_reset():
    var = ConfigVar("test", default=1)

    reset = var.set(2)
    reset.reset()

    with pytest.raises(ValueError):  # noqa: PT011
        reset.reset()


@pytest.mark.asyncio
async def test_config_var_dependency(app, client):
    var = ConfigVar("test", default=1)

    @app.get("/", dependencies=[Depends(var.dependency(2))])
    async def index():
        assert var.get() == 2

    response = await client.get("/")

    assert response.status_code == status.HTTP_200_OK
    assert var.get() == 1



from typing import Any

import pytest
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, select, true
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.orm import declarative_base, relationship

from fastapi_filters_standard import FilterField, FilterSet
from fastapi_filters_standard.ext.sqlalchemy import (
    apply_filters,
    apply_filters_and_sorting,
    apply_sorting,
    create_filters_from_orm,
    create_sorting_from_orm,
)
from fastapi_filters_standard.operators import FilterOperator

Base = declarative_base()


class Group(Base):
    __tablename__ = "groups"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    created_at = Column(DateTime)

    users = relationship("User", back_populates="group")


class User(Base):
    __tablename__ = "users"

    id = Column(Integer, primary_key=True)
    group_id = Column(ForeignKey("groups.id"))

    name = Column(String)
    age = Column(Integer)
    languages = Column(ARRAY(String))
    created_at = Column(DateTime)

    group = relationship("Group", back_populates="users")


def _compile_expr(expr: Any) -> str:
    return str(expr.compile(compile_kwargs={"literal_binds": True}))


@pytest.mark.parametrize(
    ("op", "val", "expected"),
    [
        (FilterOperator.eq, 10, User.age == 10),
        (FilterOperator.ne, 10, User.age != 10),
        (FilterOperator.gt, 10, User.age > 10),
        (FilterOperator.ge, 10, User.age >= 10),
        (FilterOperator.lt, 10, User.age < 10),
        (FilterOperator.le, 10, User.age <= 10),
        (FilterOperator.in_, [10, 20], User.age.in_([10, 20])),
        (FilterOperator.not_in, [10, 20], ~User.age.in_([10, 20])),
        (FilterOperator.is_null, True, User.age.is_(None)),
        (FilterOperator.is_null, False, User.age.isnot(None)),
    ],
    ids=[
        "eq",
        "ne",
        "gt",
        "ge",
        "lt",
        "le",
        "in",
        "not_in",
        "is_null",
        "is_not_null",
    ],
)
def test_apply_filters(op, val, expected):
    stmt = select(User)

    filtered_stmt = apply_filters(stmt, {"age": {op: val}})

    assert _compile_expr(filtered_stmt.whereclause) == _compile_expr(expected)


@pytest.mark.parametrize(
    ("op", "val", "expected"),
    [
        (FilterOperator.like, "%test%", User.name.like("%test%")),
        (FilterOperator.ilike, "%test%", User.name.ilike("%test%")),
        (FilterOperator.not_like, "%test%", ~User.name.like("%test%")),
        (FilterOperator.not_ilike, "%test%", ~User.name.ilike("%test%")),
    ],
)
def test_apply_filters_string(op, val, expected):
    stmt = select(User)

    filtered_stmt = apply_filters(stmt, {"name": {op: val}})

    assert _compile_expr(filtered_stmt.whereclause) == _compile_expr(expected)


@pytest.mark.parametrize(
    ("op", "val", "expected"),
    [
        (FilterOperator.overlap, ["en", "ua"], User.languages.overlap(["en", "ua"])),
        (
            FilterOperator.not_overlap,
            ["en", "ua"],
            ~User.languages.overlap(["en", "ua"]),
        ),
        (FilterOperator.contains, "en", User.languages.contains("en")),
        (FilterOperator.not_contains, "en", ~User.languages.contains("en")),
    ],
    ids=[
        "ov",
        "not_ov",
        "contains",
        "not_contains",
    ],
)
def test_arrays_apply_filters(op, val, expected):
    stmt = select(User)

    filtered_stmt = apply_filters(stmt, {"languages": {op: val}})

    actual_sql = filtered_stmt.whereclause.compile()
    expected_sql = expected.compile()

    assert str(actual_sql) == str(expected_sql)
    assert actual_sql.params == expected_sql.params


def test_apply_filters_with_join():
    stmt = select(User).join(Group)

    expected = (Group.name == "test") & (User.name != "John")

    filtered_stmt = apply_filters(
        stmt,
        {
            "group.name": {FilterOperator.eq: "test"},
            "user.name": {FilterOperator.ne: "John"},
        },
    )

    assert _compile_expr(filtered_stmt.whereclause) == _compile_expr(expected)


def test_apply_filters_filter_set():
    class _FilterSet(FilterSet):
        name: FilterField[str]
        age: FilterField[int]
        is_adult: FilterField[bool]

    stmt = select(User)

    _filter_set = _FilterSet(
        name={
            FilterOperator.eq: "John",
            FilterOperator.ne: "Doe",
        },
        age={
            FilterOperator.gt: 10,
            FilterOperator.lt: 20,
        },
        is_adult={
            FilterOperator.eq: True,
        },
    )
    filtered_stmt = apply_filters(
        stmt,
        _filter_set,
        additional={
            _FilterSet.is_adult: User.age >= 18,
        },
    )

    expected = (
        (User.name == "John") & (User.name != "Doe") & (User.age > 10) & (User.age < 20) & ((User.age >= 18) == true())
    )

    assert _compile_expr(filtered_stmt.whereclause) == _compile_expr(expected)


def test_unknown_operator():
    stmt = select(User)

    with pytest.raises(NotImplementedError, match=r"Operator unknown is not implemented"):
        apply_filters(stmt, {"name": {"unknown": "test"}})


def test_apply_sorting():
    stmt = select(User)

    sorted_stmt = apply_sorting(
        stmt,
        [("name", "asc", None), ("age", "desc", "bigger")],
    )

    assert _compile_expr(sorted_stmt) == _compile_expr(
        stmt.order_by(User.name.asc(), User.age.desc().nulls_first()),
    )


def test_apply_sorting_invalid_direction():
    stmt = select(User)

    with pytest.raises(ValueError, match=r"^Unknown sorting direction .*$"):
        apply_sorting(stmt, [("name", "invalid", None)])


def test_apply_filtering_sorting():
    stmt = select(User)

    result_stmt = apply_filters_and_sorting(
        stmt,
        {
            "name": {FilterOperator.eq: "test"},
            "age": {FilterOperator.gt: 10},
        },
        [("name", "asc", None), ("age", "desc", None)],
    )

    assert _compile_expr(result_stmt) == _compile_expr(
        stmt.where(User.name == "test").where(User.age > 10).order_by(User.name.asc(), User.age.desc()),
    )


def test_create_filters_from_orm():
    resolver = create_filters_from_orm(User, include={"age"})

    assert resolver.__defs__ == {
        "age": ("age", FilterOperator.eq),
        "age__eq": ("age", FilterOperator.eq),
        "age__gte": ("age", FilterOperator.ge),
        "age__gt": ("age", FilterOperator.gt),
        "age__in": ("age", FilterOperator.in_),
        "age__isnull": ("age", FilterOperator.is_null),
        "age__lte": ("age", FilterOperator.le),
        "age__lt": ("age", FilterOperator.lt),
        "age__ne": ("age", FilterOperator.ne),
        "age__not_in": ("age", FilterOperator.not_in),
    }

    resolver = create_filters_from_orm(
        User,
        exclude={"id", "name", "languages", "created_at"},
    )

    assert resolver.__defs__ == {
        "age": ("age", FilterOperator.eq),
        "age__eq": ("age", FilterOperator.eq),
        "age__gte": ("age", FilterOperator.ge),
        "age__gt": ("age", FilterOperator.gt),
        "age__in": ("age", FilterOperator.in_),
        "age__isnull": ("age", FilterOperator.is_null),
        "age__lte": ("age", FilterOperator.le),
        "age__lt": ("age", FilterOperator.lt),
        "age__ne": ("age", FilterOperator.ne),
        "age__not_in": ("age", FilterOperator.not_in),
    }

    resolver = create_filters_from_orm(User, include={"languages"})

    assert resolver.__defs__ == {
        "languages": ("languages", FilterOperator.eq),
        "languages__isnull": ("languages", FilterOperator.is_null),
        "languages__overlap": ("languages", FilterOperator.overlap),
        "languages__not_overlap": ("languages", FilterOperator.not_overlap),
        "languages__contains": ("languages", FilterOperator.contains),
        "languages__not_contains": ("languages", FilterOperator.not_contains),
    }


def test_create_sorting_from_orm():
    resolver = create_sorting_from_orm(User)

    assert resolver.__defs__ == {
        "-id": ("id", "desc", None),
        "+id": ("id", "asc", None),
        "+age": ("age", "asc", None),
        "-age": ("age", "desc", None),
        "-created_at": ("created_at", "desc", None),
        "+created_at": ("created_at", "asc", None),
        "-languages": ("languages", "desc", None),
        "+languages": ("languages", "asc", None),
        "-name": ("name", "desc", None),
        "+name": ("name", "asc", None),
    }



import inspect
from collections.abc import Awaitable, Callable, Mapping, Sequence
from dataclasses import InitVar, asdict, dataclass, replace
from typing import (
    Any,
    ClassVar,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

from fastapi import Depends
from typing_extensions import Self, dataclass_transform

from .fields import FilterField
from .filters import create_filters
from .op import FilterOp
from .types import AbstractFilterOperator, FiltersResolver, FilterValues

T_co = TypeVar("T_co", covariant=True)


class FilterSetMeta(type):
    def __init__(
        cls,
        name: str,
        bases: Any,
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> None:
        super().__init__(name, bases, namespace, **kwargs)

        hints = get_type_hints(cls, include_extras=True)
        specs = {key: get_args(value)[0] for key, value in hints.items() if get_origin(value) is FilterField}

        cls.__filters__: dict[str, FilterField[Any]] = {}
        for base in bases:
            cls.__filters__.update(getattr(base, "__filters__", {}))

        for key, value in specs.items():
            try:
                filter_field = getattr(cls, key)
            except AttributeError:
                filter_field = FilterField(value)
                setattr(cls, key, filter_field)

            if filter_field.type is None:
                filter_field = replace(filter_field, type=value)

            filter_field.__set_name__(cls, key)
            cls.__filters__[key] = filter_field

        for key in cls.__filters__:  # used to add default value for dataclass
            setattr(cls, key, None)

        d_cls = dataclass(
            cast(type[Any], cls),
            kw_only=True,
        )

        for key, value in cls.__filters__.items():
            setattr(d_cls, key, value)

        try:
            _ = FilterSet
        except NameError:
            return

        d_cls.__signature__ = inspect.Signature(
            parameters=[
                inspect.Parameter(
                    name="__values__",
                    kind=inspect.Parameter.KEYWORD_ONLY,
                    default=Depends(_filters_from_set(d_cls)),
                    annotation=FilterValues,
                )
            ],
            return_annotation=FilterSet,
        )


@dataclass_transform(
    field_specifiers=(FilterField,),
)
class FilterSet(metaclass=FilterSetMeta):
    __values__: InitVar[FilterValues | None] = None  # type: ignore[assignment]

    __filters__: ClassVar[dict[str, FilterField[Any]]]

    @classmethod
    def create(
        cls,
        **kwargs: Mapping[AbstractFilterOperator, Sequence[Any] | bool | Any],
    ) -> Self:
        return cls(**kwargs)

    def __post_init__(
        self,
        __values__: FilterValues | None,
    ) -> None:
        for key, value in (__values__ or {}).items():
            setattr(self, key, value)

        for key in asdict(self):  # type: ignore[call-overload]
            # auto replace uninitialized fields with empty dict
            if getattr(self, key) is None:
                setattr(self, key, {})

        self.init_filter_set()

    def __bool__(self) -> bool:
        return bool(self.filter_values)

    def init_filter_set(self) -> None:
        pass

    @classmethod
    def from_ops(cls, *ops: FilterOp[Any]) -> Self:
        values: FilterValues = {k: {} for k in cls.__filters__}
        for op in ops:
            if op.operator in values[op.name]:
                raise ValueError(f"Duplicate operator {op.operator.name} for {op.name}")

            values[op.name][op.operator] = op.value

        return cls(**{k: v or None for k, v in values.items()})

    @classmethod
    def create_from_resolver(
        cls,
        resolver: FiltersResolver,
        *,
        name: str | None = None,
    ) -> type[Self]:
        anns = {k: FilterField[f.type] for k, f in resolver.__filters__.items()}  # type: ignore[name-defined]

        return type(
            name or "FilterSet",
            (cls,),
            {**resolver.__filters__, "__annotations__": anns},
        )

    @property
    def filter_values(self) -> FilterValues:
        return {key: val for key in self.__filters__ if (val := getattr(self, key))}

    def subset(self, *fields: FilterField[Any] | str) -> Self:
        names = {name for f in fields if (name := f.name if isinstance(f, FilterField) else f)}
        return self.create(
            **{k: v for k, v in self.filter_values.items() if k in names},
        )

    def extract(
        self,
        *fields: FilterField[Any] | str | type["FilterSet"],
        strict: bool = False,
    ) -> Self:
        names = set()
        for f in fields:
            if isinstance(f, FilterField):
                assert f.name is not None, "FilterField name is not set"
                names.add(f.name)
            elif isinstance(f, type) and issubclass(f, FilterSet):
                names |= f.__filters__.keys()
            else:
                names.add(f)

        if not strict:
            names &= self.filter_values.keys()

        attrs = {}
        for name in names:
            if val := getattr(self, name):
                attrs[name] = {**val}
                val.clear()

        return self.create(**attrs)


TFiltersSet = TypeVar("TFiltersSet", bound=FilterSet)


def _filters_from_set(
    filters_set: type[TFiltersSet],
) -> FiltersResolver:
    return create_filters(
        **{k: v for k, v in filters_set.__filters__.items() if not v.internal},  # type: ignore[arg-type]
    )


def create_filters_from_set(
    filters_set: type[TFiltersSet],
) -> Callable[..., Awaitable[TFiltersSet]]:
    filters_dep = _filters_from_set(filters_set)

    async def resolver(values: FilterValues = Depends(filters_dep)) -> TFiltersSet:
        return filters_set.create(**values)

    for attr in ("__filters__", "__model__", "__defs__"):
        setattr(resolver, attr, getattr(filters_dep, attr))

    return resolver


__all__ = [
    "FilterSet",
    "create_filters_from_set",
]



import threading
from collections.abc import Awaitable, Sequence

import pytest
from fastapi import Depends

from fastapi_filters_standard.utils import (
    async_safe,
    is_optional,
    is_seq,
    unwrap_optional_type,
    unwrap_seq_type,
    unwrap_type,
)


@pytest.mark.asyncio
async def test_async_safe(app, client):
    def foo() -> int:
        return threading.get_ident()

    @app.get("/")
    async def route(not_safe: int = Depends(foo), safe: int = Depends(async_safe(foo))):
        return [not_safe, safe]

    res = await client.get("/")
    a, b = res.json()

    assert a != threading.get_ident()
    assert b == threading.get_ident()


def test_is_seq():
    assert is_seq(list)
    assert is_seq(list[int])
    assert is_seq(tuple[int, ...])
    assert is_seq(Sequence[int])

    assert not is_seq(None)
    assert not is_seq(1)
    assert not is_seq(Awaitable[int])


def test_is_optional():
    assert is_optional(int | None)
    assert is_optional(list[int] | None)
    assert is_optional(tuple[int, ...] | None)
    assert is_optional(Sequence[int] | None)
    assert is_optional(float | int | None)

    assert not is_optional(None)
    assert not is_optional(1)
    assert not is_optional(Awaitable[int])


def test_is_optional_union_operator():
    assert is_optional(int | None)
    assert is_optional(int | float | None)

    assert not is_optional(int | float)


def test_unwrap_type():
    assert unwrap_type(int | None) is int
    assert unwrap_type(int | float | None) == int | float
    assert unwrap_type(list[int]) is int
    assert unwrap_type(tuple[int, ...]) is int
    assert unwrap_type(tuple[int, float]) == int | float
    assert unwrap_type(Sequence[int]) is int
    assert unwrap_type(int) is int
    assert unwrap_type(None) is None


def test_unwrap_type_generic_aliases():
    assert unwrap_type(list[int]) is int
    assert unwrap_type(tuple[int, ...]) is int
    assert unwrap_type(tuple[int, float]) == int | float


def test_unwrap_type_union_operator():
    assert unwrap_type(int | None) is int
    assert unwrap_type(int | float | None) == int | float
    assert unwrap_type(int | float) == int | float


def test_unwrap_optional_type():
    assert unwrap_optional_type(int | None) is int
    assert unwrap_optional_type(list[int] | None) == list[int]
    assert unwrap_optional_type(int | float | None) == int | float

    with pytest.raises(TypeError):
        assert unwrap_optional_type(None) is None

    with pytest.raises(TypeError):
        assert unwrap_optional_type(int) is int


def test_unwrap_seq_type():
    assert unwrap_seq_type(list[int]) is int
    assert unwrap_seq_type(tuple[int, ...]) is int
    assert unwrap_seq_type(Sequence[int]) is int

    with pytest.raises(TypeError):
        assert unwrap_seq_type(None) is None

    with pytest.raises(TypeError):
        assert unwrap_seq_type(int) is int



from .filters import alias_generator_config as alias_generator
from .operators import (
    disabled_filters_config as disabled_filters,
)
from .operators import (
    filter_operators_generator_config as filter_operators_generator,
)

__all__ = [
    "alias_generator",
    "disabled_filters",
    "filter_operators_generator",
]



from enum import Enum
from typing import TypeAlias

from typing_extensions import reveal_type

from fastapi_filters_standard import FilterField


class MyEnum(str, Enum):
    A = "a"
    B = "b"
    C = "c"


_T: TypeAlias = MyEnum
val: _T = MyEnum.A

field: FilterField[_T] = FilterField()

# cmp operators
reveal_type(field == val)
reveal_type(field != val)
reveal_type(field > val)
reveal_type(field >= val)
reveal_type(field < val)
reveal_type(field <= val)

# cmp operators (funcs)
reveal_type(field.eq(val))
reveal_type(field.ne(val))
reveal_type(field.gt(val))
reveal_type(field.ge(val))
reveal_type(field.lt(val))
reveal_type(field.le(val))

# in/not in
reveal_type(field.in_([val]))
reveal_type(field >> [val])
reveal_type(field.not_in([val]))

# is null operators
reveal_type(field.is_null())
reveal_type(field.is_null(True))
reveal_type(field.is_null(False))

# str operators
reveal_type(field.like("1"))
reveal_type(field.not_like("1"))
reveal_type(field.ilike("1"))
reveal_type(field.not_ilike("1"))

# array operators
reveal_type(field.overlaps(val))
reveal_type(field.not_overlaps(val))
reveal_type(field.contains(val))
reveal_type(field.not_contains(val))

# output:
"""
tests/mypy/enum_op.py:21:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[enum.Enum]"
tests/mypy/enum_op.py:22:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[enum.Enum]"
tests/mypy/enum_op.py:23:13: note: Revealed type is "None"
tests/mypy/enum_op.py:24:13: note: Revealed type is "None"
tests/mypy/enum_op.py:25:13: note: Revealed type is "None"
tests/mypy/enum_op.py:26:13: note: Revealed type is "None"
tests/mypy/enum_op.py:29:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[enum.Enum]"
tests/mypy/enum_op.py:30:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[enum.Enum]"
tests/mypy/enum_op.py:31:13: note: Revealed type is "None"
tests/mypy/enum_op.py:32:13: note: Revealed type is "None"
tests/mypy/enum_op.py:33:13: note: Revealed type is "None"
tests/mypy/enum_op.py:34:13: note: Revealed type is "None"
tests/mypy/enum_op.py:37:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[typing.Sequence[enum.Enum]]"
tests/mypy/enum_op.py:38:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[typing.Sequence[enum.Enum]]"
tests/mypy/enum_op.py:39:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[typing.Sequence[enum.Enum]]"
tests/mypy/enum_op.py:42:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[builtins.bool]"
tests/mypy/enum_op.py:43:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[builtins.bool]"
tests/mypy/enum_op.py:44:13: note: Revealed type is "fastapi_filters_standard.op.FilterOp[builtins.bool]"
tests/mypy/enum_op.py:47:13: note: Revealed type is "Any"
tests/mypy/enum_op.py:48:13: note: Revealed type is "Any"
tests/mypy/enum_op.py:49:13: note: Revealed type is "Any"
tests/mypy/enum_op.py:50:13: note: Revealed type is "Any"
tests/mypy/enum_op.py:53:13: note: Revealed type is "None"
tests/mypy/enum_op.py:54:13: note: Revealed type is "None"
tests/mypy/enum_op.py:55:13: note: Revealed type is "None"
tests/mypy/enum_op.py:56:13: note: Revealed type is "None"
Success: no issues found in 1 source file
"""



import re
from collections.abc import Mapping
from dataclasses import dataclass
from functools import cache, cached_property
from typing import TYPE_CHECKING, Any, NamedTuple, TypeAlias, cast

from sqlalchemy import Dialect, column, make_url, select
from sqlalchemy.sql import ClauseElement
from sqlalchemy.sql.compiler import StrSQLCompiler
from sqlalchemy.sql.type_api import TypeEngine

from fastapi_filters_standard import FilterSet, FilterValues
from fastapi_filters_standard.config import ConfigVar
from fastapi_filters_standard.types import SortingValues

from .sqlalchemy import DEFAULT_FILTERS, SORT_FUNCS, SORT_NULLS_FUNCS

_Dialect: TypeAlias = str | None
_SQLType: TypeAlias = TypeEngine[Any] | type[TypeEngine[Any]]

default_dialect: ConfigVar[_Dialect] = ConfigVar(
    name="default_dialect",
    default=None,
)
default_compile_kwargs: ConfigVar[_Dialect] = ConfigVar(
    name="default_compile_kwargs",
    default=None,
)


@cache
def _get_dialect(dialect: str) -> Dialect:
    return make_url(f"{dialect}://").get_dialect()()


if TYPE_CHECKING:

    class _BaseCompiledStatement(NamedTuple):
        stmt: str
        args: tuple[Any, ...]
else:

    class _BaseCompiledStatement:
        stmt: str
        args: tuple[Any, ...]

        def __getitem__(self, item):
            if item == 0:
                return self.stmt
            if item == 1:
                return self.args

            raise IndexError("Index out of range for CompiledStatement")

        def __len__(self):
            return 2


def _adjust_args_pos_in_statement(
    num_char: str,
    stmt: str,
    arg_start: int,
) -> str:
    regex = re.compile(rf"\{num_char}(\d+)")

    return regex.sub(
        lambda match: f"{num_char}{int(match.group(1)) + arg_start - 1}",
        stmt,
    )


@dataclass(kw_only=True)
class CompiledStatement(_BaseCompiledStatement):  # type: ignore[misc]
    arg_start: int = 1
    _compiled: StrSQLCompiler

    @property
    def start(self) -> int:
        return self.arg_start + self.end - len(self.args)

    @property
    def end(self) -> int:
        return self.arg_start + len(self.args)

    @property
    def nargs(self) -> int:
        return len(self.args)

    @cached_property
    def stmt(self) -> str:
        if self.is_positional and self.arg_start != 1:
            return _adjust_args_pos_in_statement(
                num_char=self._compiled._numeric_binds_identifier_char,
                stmt=self._compiled.string,
                arg_start=self.arg_start,
            )

        return self._compiled.string

    @property
    def is_positional(self) -> bool:
        return self._compiled.positional

    @property
    def args(self) -> tuple[Any, ...]:
        params = self._compiled.params

        if self.is_positional:
            return tuple(params[name] for name in self._compiled.positiontup or ())

        return tuple(params.values())

    @property
    def params(self) -> Mapping[str, Any]:
        return cast(Mapping[str, Any], self._compiled.params)


def _compile_sql(
    stmt: ClauseElement,
    *,
    dialect: _Dialect | None = None,
    arg_start: int | None = None,
) -> CompiledStatement:
    dialect = dialect or default_dialect.get()
    sa_dialect: Dialect | None = _get_dialect(dialect) if dialect else None

    if arg_start is None:
        arg_start = 1

    compiled = stmt.compile(
        dialect=sa_dialect,
        compile_kwargs=default_compile_kwargs.get() or {},
    )

    return CompiledStatement(  # type: ignore[call-arg]
        arg_start=arg_start,
        _compiled=cast(StrSQLCompiler, compiled),
    )


def apply_filters(
    filters: FilterValues | FilterSet,
    *,
    remapping: Mapping[str, str] | None = None,
    types: Mapping[str, _SQLType] | None = None,
    dialect: _Dialect | None = None,
    arg_start: int | None = None,
) -> CompiledStatement | None:
    types = types or {}
    remapping = remapping or {}
    if isinstance(filters, FilterSet):
        filters = filters.filter_values

    if not filters:
        return None

    stmt = select(1)
    for field, field_filters in filters.items():
        field = remapping.get(field, field)

        for op, val in field_filters.items():
            if (cond := DEFAULT_FILTERS.get(op)) is not None:
                col = column(field, type_=types.get(field))
                stmt = stmt.where(cond(col, val))
            else:
                raise NotImplementedError(f"Operator {op} is not implemented")

    return _compile_sql(
        cast(ClauseElement, stmt.whereclause),
        dialect=dialect,
        arg_start=arg_start,
    )


def apply_sorting(
    sorting: SortingValues,
    *,
    remapping: Mapping[str, str] | None = None,
    dialect: _Dialect | None = None,
    types: Mapping[str, _SQLType] | None = None,
    arg_start: int | None = None,
) -> CompiledStatement | None:
    types = types or {}
    remapping = remapping or {}

    if not sorting:
        return None

    stmt = select(1)
    for field, direction, nulls in sorting:
        field = remapping.get(field, field)
        col = column(field, type_=types.get(field))

        if sort_func := SORT_FUNCS.get(direction):
            sort_expr = sort_func(col)

            if nulls is not None:
                sort_expr = SORT_NULLS_FUNCS[(direction, nulls)](sort_expr)

            stmt = stmt.order_by(sort_expr)
        else:
            raise ValueError(f"Unknown sorting direction {direction}")

    return _compile_sql(
        stmt._order_by_clause,
        dialect=dialect,
        arg_start=arg_start,
    )


def apply_filters_and_sorting(
    filters: FilterValues | FilterSet,
    sorting: SortingValues,
    *,
    dialect: _Dialect | None = None,
    remapping: Mapping[str, str] | None = None,
    types: Mapping[str, _SQLType] | None = None,
    arg_start: int | None = None,
) -> tuple[CompiledStatement | None, CompiledStatement | None]:
    filters_res = apply_filters(
        filters,
        arg_start=arg_start,
        remapping=remapping,
        types=types,
        dialect=dialect,
    )

    if filters_res and filters_res.is_positional:
        arg_start = filters_res.end

    sorting_res = apply_sorting(
        sorting,
        arg_start=arg_start,
        remapping=remapping,
        types=types,
        dialect=dialect,
    )

    return filters_res, sorting_res


__all__ = [
    "CompiledStatement",
    "apply_filters",
    "apply_filters_and_sorting",
    "apply_sorting",
    "default_compile_kwargs",
    "default_dialect",
]



from collections.abc import Callable
from enum import Enum
from typing import (
    Any,
    Literal,
    Protocol,
    TypeAlias,
)

from .fields import FilterField

AbstractFilterOperator: TypeAlias = Enum
FilterAliasGenerator: TypeAlias = Callable[
    [str, AbstractFilterOperator, str | None],
    str,
]
FilterPlace: TypeAlias = Callable[..., Any]
FilterFieldDef: TypeAlias = type[Any] | FilterField[Any]
FilterValues: TypeAlias = dict[str, dict[AbstractFilterOperator, Any]]


class FiltersResolver(Protocol):
    __model__: type[Any]
    __defs__: dict[str, tuple[str, AbstractFilterOperator]]
    __filters__: dict[str, FilterField[Any]]

    async def __call__(self, _: Any, /) -> FilterValues:  # pragma: no cover
        pass


SortingDirection: TypeAlias = Literal["asc", "desc"]
SortingNulls: TypeAlias = Literal["bigger", "smaller"] | None
SortingValues: TypeAlias = list[tuple[str, SortingDirection, SortingNulls]]


class SortingResolver(Protocol):
    __tp__: Any
    __defs__: dict[str, tuple[str, SortingDirection, SortingNulls]]

    async def __call__(self, _: Any, /) -> SortingValues:  # pragma: no cover
        pass


__all__ = [
    "AbstractFilterOperator",
    "FilterAliasGenerator",
    "FilterFieldDef",
    "FilterPlace",
    "FilterValues",
    "FiltersResolver",
    "SortingDirection",
    "SortingNulls",
    "SortingResolver",
    "SortingValues",
]



from collections.abc import Container
from typing import Annotated, Literal, cast

from fastapi import Query
from pydantic import BaseModel

from .schemas import CSVList
from .types import FilterPlace, SortingNulls, SortingResolver, SortingValues
from .utils import fields_include_exclude, is_complex_field


def create_sorting_from_model(
    model: type[BaseModel],
    *,
    default: str | None = None,
    in_: FilterPlace | None = None,
    include: Container[str] | None = None,
    exclude: Container[str] | None = None,
) -> SortingResolver:
    checker = fields_include_exclude(model.model_fields, include, exclude)

    return create_sorting(
        *[name for name, field in model.model_fields.items() if checker(name) and not is_complex_field(field)],
        in_=in_,
        default=default,
    )


def create_sorting(
    *fields: str | tuple[str, SortingNulls],
    in_: FilterPlace | None = None,
    default: str | list[str] | None = None,
    alias: str | None = None,
) -> SortingResolver:
    if in_ is None:
        in_ = Query

    normalized_fields = [(f, None) if isinstance(f, str) else f for f in fields]

    defs = {f"{d}{f}": (f, v, n) for (v, d) in (("asc", "+"), ("desc", "-")) for f, n in normalized_fields}
    tp = Literal[tuple(defs)]  # type: ignore[valid-type]

    default = [default] if isinstance(default, str) else default

    if default and (diff := {*default} - {*defs}):
        raise ValueError(
            f"Default sort field {','.join(diff)} is not in {','.join(f for f, _ in normalized_fields)}",
        )

    async def _get_sorters(
        sort: Annotated[CSVList[tp], in_(alias=alias, explode=False)] = default,  # type: ignore[valid-type,assignment]
    ) -> SortingValues:
        return cast(SortingValues, [defs[f] for f in sort or ()])

    _get_sorters.__tp__ = tp  # type: ignore[attr-defined]
    _get_sorters.__defs__ = defs  # type: ignore[attr-defined]

    return cast(SortingResolver, _get_sorters)


__all__ = [
    "create_sorting",
    "create_sorting_from_model",
]



import types
from collections.abc import Awaitable, Callable, Container, Iterable, Sequence
from functools import wraps
from typing import (
    Annotated,
    Any,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from fastapi.dependencies.utils import lenient_issubclass
from pydantic.fields import FieldInfo
from typing_extensions import ParamSpec

try:
    from fastapi._compat.shared import field_annotation_is_complex
except ImportError:
    from fastapi._compat import field_annotation_is_complex  # type: ignore[attr-defined,no-redef]

P = ParamSpec("P")
T = TypeVar("T")


def async_safe(f: Callable[P, T]) -> Callable[P, Awaitable[T]]:
    @wraps(f)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return f(*args, **kwargs)

    return wrapper


def is_none_type(tp: Any) -> bool:
    return tp in (type(None), None, None)


def is_union(tp: Any) -> bool:
    return tp in (Union, types.UnionType) or get_origin(tp) in (Union, types.UnionType)


def is_optional(tp: Any) -> bool:
    return is_union(get_origin(tp)) and any(is_none_type(arg) for arg in get_args(tp))


def is_seq(tp: Any) -> bool:
    return not lenient_issubclass(tp, (str, bytes)) and (
        lenient_issubclass(tp, Sequence) or lenient_issubclass(get_origin(tp), Sequence)
    )


def _create_union(*args: Any, exclude_none: bool = True) -> Any:
    args = tuple(arg for arg in args if arg is not ...)
    if exclude_none:
        args = tuple(arg for arg in args if not is_none_type(arg))

    return Union[args]


def unwrap_optional_type(tp: Any) -> Any:
    if not is_optional(tp):
        raise TypeError(f"Expected optional type, got {tp}")

    return _create_union(*get_args(tp), exclude_none=True)


def unwrap_seq_type(tp: Any) -> Any:
    if not is_seq(tp):
        raise TypeError(f"Expected sequence type, got {tp}")

    return _create_union(*(get_args(tp) or (Any,)), exclude_none=False)


def unwrap_type(tp: Any) -> Any:
    if is_optional(tp):
        return unwrap_optional_type(tp)
    if is_seq(tp):
        return unwrap_seq_type(tp)

    return tp


def unwrap_annotated(tp: Any) -> Any:
    if get_origin(tp) is Annotated:
        return tp.__origin__

    return tp


def fields_include_exclude(
    fields: Iterable[str],
    include: Container[str] | None = None,
    exclude: Container[str] | None = None,
) -> Callable[[str], bool]:
    if include is None:
        include = {*fields}
    if exclude is None:
        exclude = ()

    def checker(field: str) -> bool:
        return field in include and field not in exclude

    return checker


def is_complex_field(field: FieldInfo) -> bool:
    return field_annotation_is_complex(field.annotation)


__all__ = [
    "async_safe",
    "fields_include_exclude",
    "is_complex_field",
    "is_optional",
    "is_seq",
    "lenient_issubclass",
    "unwrap_annotated",
    "unwrap_optional_type",
    "unwrap_seq_type",
    "unwrap_type",
]



from collections.abc import Callable, Mapping
from typing import Any, TypeVar, cast

from beanie import SortDirection
from beanie.odm.operators.find import BaseFindOperator
from beanie.odm.operators.find.comparison import GT, GTE, LT, LTE, NE, Eq, In, NotIn
from beanie.odm.operators.find.evaluation import RegEx
from beanie.odm.operators.find.logical import Not
from beanie.odm.queries.find import FindMany

from fastapi_filters_standard import FilterOperator, FilterSet, FilterValues
from fastapi_filters_standard.types import AbstractFilterOperator, SortingValues

DEFAULT_FILTERS: Mapping[AbstractFilterOperator, Callable[..., BaseFindOperator]] = {
    FilterOperator.eq: Eq,
    FilterOperator.ne: NE,
    FilterOperator.gt: GT,
    FilterOperator.ge: GTE,
    FilterOperator.lt: LT,
    FilterOperator.le: LTE,
    FilterOperator.like: RegEx,
    FilterOperator.not_like: lambda field, val: Not(RegEx(field, val)),
    FilterOperator.ilike: lambda field, val: RegEx(field, val, options="i"),
    FilterOperator.not_ilike: lambda field, val: Not(RegEx(field, val, options="i")),
    FilterOperator.in_: In,
    FilterOperator.not_in: NotIn,
    FilterOperator.is_null: lambda field, val: Eq(field, None) if val else NE(field, None),
    FilterOperator.contains: In,
    FilterOperator.not_contains: NotIn,
    FilterOperator.overlap: In,
    FilterOperator.not_overlap: NotIn,
}


TStmt = TypeVar("TStmt", bound=FindMany[Any])


def apply_filters(
    stmt: TStmt,
    filters: FilterValues | FilterSet,
    *,
    remapping: Mapping[str, str] | None = None,
) -> TStmt:
    remapping = remapping or {}
    if isinstance(filters, FilterSet):
        filters = filters.filter_values

    for field, field_filters in filters.items():
        field = remapping.get(field, field)

        for op, val in field_filters.items():
            if (cond := DEFAULT_FILTERS.get(op)) is not None:
                stmt = cast(TStmt, stmt.find(cond(field, val)))
            else:
                raise NotImplementedError(f"Operator {op} is not implemented")

    return stmt


def apply_sorting(
    stmt: TStmt,
    sorting: SortingValues,
    remapping: Mapping[str, str] | None = None,
) -> TStmt:
    remapping = remapping or {}
    ordering = []
    for field, direction, _ in sorting:
        field = remapping.get(field, field)

        if direction == "asc":
            ordering.append((field, SortDirection.ASCENDING))
        else:
            ordering.append((field, SortDirection.DESCENDING))

    if ordering:
        stmt = cast(TStmt, stmt.sort(*ordering))

    return stmt


def apply_filters_and_sorting(
    stmt: TStmt,
    filters: FilterValues | FilterSet,
    sorting: SortingValues,
    *,
    remapping: Mapping[str, str] | None = None,
) -> TStmt:
    stmt = apply_filters(stmt, filters, remapping=remapping)
    return apply_sorting(stmt, sorting, remapping=remapping)


__all__ = [
    "apply_filters",
    "apply_filters_and_sorting",
    "apply_sorting",
]



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Protocol,
    TypeVar,
    overload,
)

from .operators import FilterOperator

if TYPE_CHECKING:
    from .types import AbstractFilterOperator


TVal_co = TypeVar("TVal_co", covariant=True)


@dataclass(frozen=True)
class FilterOp(Generic[TVal_co]):
    name: str
    operator: AbstractFilterOperator
    value: TVal_co


def _simple_op(
    operator: AbstractFilterOperator,
) -> Callable[[_HasNameAndOperatorsProtocol, Any], FilterOp[Any]]:
    def func(self: _HasNameAndOperatorsProtocol, value: Any) -> FilterOp[Any]:
        self._check_op(operator)
        return FilterOp(self.name, operator, value)  # type: ignore[arg-type]

    func.__name__ = operator.name
    return func


class _HasNameAndOperatorsProtocol(Protocol):
    name: str | None
    operators: list[AbstractFilterOperator] | None

    @abstractmethod
    def _check_op(
        self: _HasNameAndOperatorsProtocol,
        operator: AbstractFilterOperator,
    ) -> None:
        pass


T_co = TypeVar("T_co", covariant=True)
TArg_co = TypeVar("TArg_co", covariant=True)


class FilterOpBuilder(Generic[T_co]):  # noqa: PLW1641
    def _check_op(
        self: _HasNameAndOperatorsProtocol,
        operator: AbstractFilterOperator,
    ) -> None:
        assert self.operators is not None, "FilterField has no operators"
        assert self.name is not None, "FilterField has no name"

        assert operator in self.operators, f"Operator {operator} is not allowed for field {self.name!r}"

    def __call__(
        self: _HasNameAndOperatorsProtocol,
        op: AbstractFilterOperator,
        value: T_co,  # type: ignore[misc]
    ) -> FilterOp[T_co]:
        self._check_op(op)
        return FilterOp(self.name, op, value)  # type: ignore[arg-type]

    if TYPE_CHECKING:

        @overload
        def __scalar_method_scalar_arg__(
            self: FilterOpBuilder[Enum],
            value: Enum,
            /,
        ) -> FilterOp[Enum]:
            pass

        @overload  # we need special overload case for str because it is a Sequence of strings
        def __scalar_method_scalar_arg__(
            self: FilterOpBuilder[str],
            value: str,
            /,
        ) -> FilterOp[str]:
            pass

        @overload
        def __scalar_method_scalar_arg__(  # type: ignore[overload-overlap]
            self: FilterOpBuilder[Sequence[TArg_co]],
            value: Any,
            /,
        ) -> None:
            pass

        @overload
        def __scalar_method_scalar_arg__(
            self: FilterOpBuilder[TArg_co],
            value: TArg_co,  # type: ignore[misc]
            /,
        ) -> FilterOp[TArg_co]:
            pass

        def __scalar_method_scalar_arg__(self, value: Any, /) -> Any:
            pass

        @overload
        def __non_scalar_method_non_scalar_arg__(
            self: FilterOpBuilder[str],
            value: Any,
            /,
        ) -> None:
            pass

        @overload
        def __non_scalar_method_non_scalar_arg__(
            self: FilterOpBuilder[Sequence[TArg_co]],
            value: Sequence[TArg_co],
            /,
        ) -> FilterOp[Sequence[TArg_co]]:
            pass

        @overload
        def __non_scalar_method_non_scalar_arg__(
            self: FilterOpBuilder[TArg_co],
            value: Any,
            /,
        ) -> None:
            pass

        def __non_scalar_method_non_scalar_arg__(self, value: Any, /) -> Any:
            pass

        @overload
        def __eq__(self: FilterOpBuilder[Enum], value: Enum, /) -> FilterOp[Enum]:  # type: ignore[overload-overlap]
            pass

        @overload
        def __eq__(self: FilterOpBuilder[str], value: str, /) -> FilterOp[str]:  # type: ignore[overload-overlap]
            pass

        @overload
        def __eq__(self: FilterOpBuilder[Sequence[TArg_co]], value: Any, /) -> None:  # type: ignore[overload-overlap]
            pass

        @overload
        def __eq__(self: FilterOpBuilder[TArg_co], value: TArg_co, /) -> FilterOp[TArg_co]:  # type: ignore[misc]
            pass

        @overload
        def __eq__(self: object, value: object, /) -> bool:
            pass

        def __eq__(self, value: Any, /) -> Any:
            pass

        __ne__ = __eq__

        @overload
        def __lt__(self: FilterOpBuilder[bool], other: Any, /) -> None:  # type: ignore[overload-overlap]
            pass

        @overload
        def __lt__(self: FilterOpBuilder[Enum], other: Any, /) -> None:  # type: ignore[overload-overlap]
            pass

        @overload
        def __lt__(  # type: ignore[overload-overlap]
            self: FilterOpBuilder[Sequence[TArg_co]],
            other: Any,
            /,
        ) -> None:
            pass

        @overload
        def __lt__(self: FilterOpBuilder[TArg_co], other: TArg_co, /) -> FilterOp[TArg_co]:  # type: ignore[misc]
            pass

        def __lt__(self, other: Any) -> Any:
            pass

        __le__ = __lt__
        __gt__ = __lt__
        __ge__ = __lt__

        eq = __eq__
        ne = __eq__

        lt = __lt__
        le = __lt__
        gt = __lt__
        ge = __lt__

        @overload
        def __rshift__(
            self: FilterOpBuilder[Enum],
            other: Any,
            /,
        ) -> FilterOp[Sequence[Enum]]:
            pass

        @overload
        def __rshift__(self: FilterOpBuilder[bool], other: Any, /) -> None:
            pass

        @overload
        def __rshift__(self: FilterOpBuilder[Sequence[TArg_co]], other: Any, /) -> None:
            pass

        @overload
        def __rshift__(
            self: FilterOpBuilder[TArg_co],
            other: Sequence[TArg_co],
            /,
        ) -> FilterOp[Sequence[TArg_co]]:
            pass

        # >> as in operator
        def __rshift__(self, other: Any) -> Any:
            pass

        in_ = __rshift__
        not_in = __rshift__

        @overload
        def like(self: FilterOpBuilder[Enum], value: str, /) -> Any:
            pass

        @overload
        def like(self: FilterOpBuilder[str], value: str, /) -> FilterOp[str]:
            pass

        @overload
        def like(self: FilterOpBuilder[Any], value: Any, /) -> None:
            pass

        def like(self, value: Any, /) -> Any:
            pass

        not_like = like
        ilike = like
        not_ilike = like

        overlaps = __non_scalar_method_non_scalar_arg__
        not_overlaps = __non_scalar_method_non_scalar_arg__

        contains = __non_scalar_method_non_scalar_arg__
        not_contains = __non_scalar_method_non_scalar_arg__

        def is_null(self, value: bool = ..., /) -> FilterOp[bool]:
            pass

    else:
        __eq__ = _simple_op(FilterOperator.eq)
        __ne__ = _simple_op(FilterOperator.ne)
        __lt__ = _simple_op(FilterOperator.lt)
        __le__ = _simple_op(FilterOperator.le)
        __gt__ = _simple_op(FilterOperator.gt)
        __ge__ = _simple_op(FilterOperator.ge)

        eq = __eq__
        ne = __ne__
        lt = __lt__
        le = __le__
        gt = __gt__
        ge = __ge__

        in_ = _simple_op(FilterOperator.in_)
        not_in = _simple_op(FilterOperator.not_in)

        __rshift__ = in_  # >> as in operator

        like = _simple_op(FilterOperator.like)
        not_like = _simple_op(FilterOperator.not_like)
        ilike = _simple_op(FilterOperator.ilike)
        not_ilike = _simple_op(FilterOperator.not_ilike)

        overlaps = _simple_op(FilterOperator.overlap)
        not_overlaps = _simple_op(FilterOperator.not_overlap)
        contains = _simple_op(FilterOperator.contains)
        not_contains = _simple_op(FilterOperator.not_contains)

        def is_null(
            self: _HasNameAndOperatorsProtocol,
            val: bool = True,
            /,
        ) -> FilterOp[bool]:
            self._check_op(FilterOperator.is_null)
            return FilterOp(self.name, FilterOperator.is_null, val)


__all__ = [
    "FilterOp",
    "FilterOpBuilder",
]



from typing_extensions import reveal_type

from fastapi_filters_standard import FilterField, FilterSet


class MyFilterSet(FilterSet):
    a: FilterField[int]
    b: FilterField[float]
    c: FilterField[str]
    d: FilterField[bool]


reveal_type(MyFilterSet.a)
reveal_type(MyFilterSet.b)
reveal_type(MyFilterSet.c)
reveal_type(MyFilterSet.d)

filter_set = MyFilterSet.create()

reveal_type(filter_set.a)
reveal_type(filter_set.b)
reveal_type(filter_set.c)
reveal_type(filter_set.d)

# output:
"""
tests/mypy/field_set_decl.py:13:13: note: Revealed type is "fastapi_filters_standard.fields.FilterField[builtins.int]"
tests/mypy/field_set_decl.py:14:13: note: Revealed type is "fastapi_filters_standard.fields.FilterField[builtins.float]"
tests/mypy/field_set_decl.py:15:13: note: Revealed type is "fastapi_filters_standard.fields.FilterField[builtins.str]"
tests/mypy/field_set_decl.py:16:13: note: Revealed type is "fastapi_filters_standard.fields.FilterField[builtins.bool]"
tests/mypy/field_set_decl.py:20:13: note: Revealed type is "builtins.dict[enum.Enum, builtins.int | typing.Sequence[builtins.int] | builtins.bool]"
tests/mypy/field_set_decl.py:21:13: note: Revealed type is "builtins.dict[enum.Enum, builtins.float | typing.Sequence[builtins.float] | builtins.bool]"
tests/mypy/field_set_decl.py:22:13: note: Revealed type is "builtins.dict[enum.Enum, builtins.str | typing.Sequence[builtins.str] | builtins.bool]"
tests/mypy/field_set_decl.py:23:13: note: Revealed type is "builtins.dict[enum.Enum, builtins.bool | typing.Sequence[builtins.bool]]"
Success: no issues found in 1 source file
"""



from datetime import date, datetime, timedelta

import pytest

from fastapi_filters_standard.operators import (
    DEFAULT_OPERATORS,
    NUM_OPERATORS,
    SEQ_OPERATORS,
    FilterOperator,
    get_filter_operators,
)


@pytest.mark.parametrize(
    ("tp", "operators"),
    [
        *[(tp, SEQ_OPERATORS) for tp in (list[int], tuple[float, ...])],
        (bool, [FilterOperator.eq, FilterOperator.ne]),
        *[(tp, DEFAULT_OPERATORS + NUM_OPERATORS) for tp in (int, float, date, datetime, timedelta)],
        (int | None, [FilterOperator.is_null, *DEFAULT_OPERATORS, *NUM_OPERATORS]),
        (list[int], SEQ_OPERATORS),
        (tuple[float, ...], SEQ_OPERATORS),
    ],
    ids=str,
)
def test_get_default_operators(tp, operators):
    assert {*get_filter_operators(tp)} == {*operators}



from __future__ import annotations

from collections.abc import AsyncIterable, Callable
from contextvars import ContextVar, Token
from dataclasses import dataclass, field
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass
class _ContextVarReset(Generic[T]):
    val: T
    var: ContextVar[T]
    token: Token[T]
    resetted: bool = field(default=False, init=False)

    def __enter__(self) -> _ContextVarReset[T]:
        return self

    def __exit__(self, *_: object) -> None:
        self.reset(required=False)

    def reset(self, required: bool = True) -> None:
        if required and self.resetted:
            raise ValueError("ConfigVar has already been resetted")

        if not self.resetted:
            self.var.reset(self.token)
            self.resetted = True


class ConfigVar(Generic[T]):
    def __init__(self, name: str, default: T) -> None:
        self.name = name
        self.default = default
        self.var: ContextVar[T] = ContextVar(name, default=default)

    def set(self, value: T) -> _ContextVarReset[T]:
        token = self.var.set(value)
        return _ContextVarReset(value, self.var, token)

    def get(self) -> T:
        return self.var.get()

    def dependency(self, value: T) -> Callable[[], AsyncIterable[_ContextVarReset[T]]]:
        async def _dependency() -> AsyncIterable[_ContextVarReset[T]]:
            with self.set(value) as reset:
                yield reset

        return _dependency


__all__ = [
    "ConfigVar",
]



from typing import Annotated, Any, TypeAlias, TypeVar

from pydantic import BeforeValidator, GetPydanticSchema

T = TypeVar("T")


def csv_list_validator(v: Any) -> Any:
    match v:
        case str():
            return v.split(",")
        case list() if all(isinstance(item, str) for item in v):
            # If FastAPI receives multiple query parameters, join them with comma
            # This prevents duplicate parameters like ?param=1&param=2&param=3
            # Example: ["1", "2", "3"] -> "1,2,3" -> ["1", "2", "3"]
            return ",".join(v).split(",")
        case list():
            # If it's already a list of non-strings (from multiple query params), join and split
            return ",".join(str(item) for item in v).split(",")
        case _:
            return v


CSVList: TypeAlias = Annotated[
    list[T],
    BeforeValidator(csv_list_validator),
    GetPydanticSchema(
        get_pydantic_json_schema=lambda core_schema, handler: {
            **handler.resolve_ref_schema(handler(core_schema)),
            "explode": False,
        },
    ),
]

__all__ = [
    "CSVList",
]



from fastapi import FastAPI
from typing import Any


def _fix_openapi_schema(openapi_schema: dict[str, Any]) -> None:
    """Fix explode parameter in OpenAPI schema."""
    for endpoints in openapi_schema.get("paths", {}).values():
        for endpoint in endpoints.values():
            for parameter in endpoint.get("parameters", ()):
                if "explode" in parameter.get("schema", {}):
                    parameter["explode"] = parameter["schema"].pop("explode")


# Global flag to track if we've already patched
_patched = False


def _auto_fix_docs() -> None:
    """Automatically patch FastAPI to fix explode parameter."""
    global _patched
    
    if _patched:
        return
    
    try:
        import fastapi
        
        # Patch FastAPI.openapi method
        original_openapi = fastapi.FastAPI.openapi
        
        def patched_openapi(self: FastAPI) -> dict[str, Any]:
            """Patched openapi method that fixes explode parameter."""
            schema = original_openapi(self)
            _fix_openapi_schema(schema)
            return schema
        
        # Replace the method
        fastapi.FastAPI.openapi = patched_openapi  # type: ignore[assignment]
        _patched = True
    except (ImportError, AttributeError):
        # FastAPI not installed or already patched
        pass


def fix_docs(app: FastAPI) -> None:
    """Manually fix docs for a specific app (legacy function, auto-patch is enabled by default)."""
    def _fix_docs() -> None:
        openapi = app.openapi()
        _fix_openapi_schema(openapi)

    app.add_event_handler("startup", _fix_docs)


__all__ = [
    "fix_docs",
    "_auto_fix_docs",
]



from collections.abc import Iterable
from pathlib import Path
from typing import Any

import pytest

from tests.utils import run_mypy


class _MypyFailure(Exception):
    pass


class _MypyItem(pytest.Item):
    def get_expected_result(self) -> str:
        *_, expected = self.path.read_text().partition("# output:")
        return expected.strip().strip('"')

    def repr_failure(
        self,
        excinfo: pytest.ExceptionInfo[BaseException],
        style: Any = None,
    ) -> Any:
        if excinfo.errisinstance(_MypyFailure):
            return str(excinfo.value)

        return super().repr_failure(excinfo, style=style)

    def runtest(self) -> Any:
        expected = self.get_expected_result()
        passed, out = run_mypy(self.path, expected)

        if not passed:
            raise _MypyFailure(out)


class _MypyTest(pytest.File):
    def collect(self) -> Iterable[pytest.Item]:
        yield _MypyItem.from_parent(
            self,
            name=f"tests/mypy/{self.name}",
            path=self.path,
            nodeid=self.path.name,
        )


def pytest_collect_file(file_path: Path, parent: pytest.Collector):
    name = file_path.name

    if name in {"__init__.py", "conftest.py"}:
        return None

    return _MypyTest.from_parent(
        parent,
        name=name,
        path=file_path,
        nodeid=name,
    )



import pytest
from fastapi import Depends, status
from pydantic import BaseModel

from fastapi_filters_standard.sorters import create_sorting, create_sorting_from_model
from fastapi_filters_standard.types import SortingValues


@pytest.mark.asyncio
async def test_filters_as_dep(app, client):
    @app.get("/")
    async def route(
        sorting: SortingValues = Depends(create_sorting("name", "age", "created_at")),
    ) -> SortingValues:
        return sorting

    res = await client.get("/")

    assert res.status_code == status.HTTP_200_OK
    assert res.json() == []

    res = await client.get("/", params={"sort": "+name,-age"})

    assert res.status_code == status.HTTP_200_OK
    assert res.json() == [["name", "asc", None], ["age", "desc", None]]


@pytest.mark.asyncio
async def test_sorting_with_comma_separated_values(app, client):
    @app.get("/")
    async def route(
        sorting: SortingValues = Depends(create_sorting("name", "age", "created_at")),
    ) -> SortingValues:
        return sorting

    # Correct: comma-separated values in single parameter
    res = await client.get("/", params={"sort": "+name,-age,+created_at"})

    assert res.status_code == status.HTTP_200_OK
    assert res.json() == [["name", "asc", None], ["age", "desc", None], ["created_at", "asc", None]]

    # Test that repeated parameters are handled correctly
    # FastAPI may receive multiple values as a list, validator should handle it
    res = await client.get("/", params=[("sort", "+name"), ("sort", "-age"), ("sort", "+created_at")])

    # The validator should join multiple values and split them
    assert res.status_code == status.HTTP_200_OK
    assert res.json() == [["name", "asc", None], ["age", "desc", None], ["created_at", "asc", None]]


def test_create_sorting():
    model = create_sorting("name", "age", "created_at")

    assert model.__defs__ == {
        "+age": ("age", "asc", None),
        "-age": ("age", "desc", None),
        "+name": ("name", "asc", None),
        "-name": ("name", "desc", None),
        "+created_at": ("created_at", "asc", None),
        "-created_at": ("created_at", "desc", None),
    }


def test_create_sorting_from_model():
    class Group(BaseModel):
        name: str

    class User(BaseModel):
        name: str
        age: int
        created_at: str
        group: Group
        languages: list[str]

    model = create_sorting_from_model(User)

    assert model.__defs__ == {
        "+age": ("age", "asc", None),
        "-age": ("age", "desc", None),
        "+name": ("name", "asc", None),
        "-name": ("name", "desc", None),
        "+created_at": ("created_at", "asc", None),
        "-created_at": ("created_at", "desc", None),
    }


def test_create_sorting_invalid_default():
    with pytest.raises(ValueError, match=r"^Default sort field invalid is not in .*$"):
        create_sorting("name", "age", "created_at", default="invalid")



from .docs import _auto_fix_docs
from .fields import FilterField
from .filter_set import FilterSet, create_filters_from_set
from .filters import create_filters, create_filters_from_model
from .operators import FilterOperator
from .sorters import create_sorting, create_sorting_from_model
from .types import FiltersResolver, FilterValues, SortingResolver, SortingValues

# Auto-patch FastAPI to fix explode parameter
_auto_fix_docs()

__all__ = [
    "FilterField",
    "FilterOperator",
    "FilterSet",
    "FilterValues",
    "FiltersResolver",
    "SortingResolver",
    "SortingValues",
    "create_filters",
    "create_filters_from_model",
    "create_filters_from_set",
    "create_sorting",
    "create_sorting_from_model",
]



from typing import Annotated

import pytest
from fastapi import Query

from fastapi_filters_standard.schemas import CSVList


@pytest.fixture
def app(app):
    @app.get("/")
    def route(
        q1: Annotated[CSVList[int], Query()],
        q2: Annotated[list[int], Query()],
    ):
        return []

    # Auto-patch is enabled by default, no need to call fix_docs
    return app


def test_fix_docs(app, client):
    assert app.openapi()["paths"]["/"]["get"]["parameters"] == [
        {
            "in": "query",
            "name": "q1",
            "required": True,
            "schema": {"items": {"type": "integer"}, "title": "Q1", "type": "array"},
            "explode": False,
        },
        {
            "in": "query",
            "name": "q2",
            "required": True,
            "schema": {"items": {"type": "integer"}, "title": "Q2", "type": "array"},
        },
    ]



import operator
from collections.abc import Callable, Container, Iterator, Mapping
from contextlib import suppress
from typing import (
    Any,
    TypeAlias,
    TypeVar,
    cast,
)

from sqlalchemy import (
    ARRAY,
    ColumnExpressionArgument,
    asc,
    desc,
    inspect,
    nulls_first,
    nulls_last,
)
from sqlalchemy.orm import ColumnProperty
from sqlalchemy.sql.selectable import Select

from fastapi_filters_standard import FilterField, create_filters
from fastapi_filters_standard.config import ConfigVar
from fastapi_filters_standard.filter_set import FilterSet
from fastapi_filters_standard.operators import FilterOperator
from fastapi_filters_standard.sorters import create_sorting
from fastapi_filters_standard.types import (
    AbstractFilterOperator,
    FilterAliasGenerator,
    FilterFieldDef,
    FilterPlace,
    FiltersResolver,
    FilterValues,
    SortingDirection,
    SortingNulls,
    SortingResolver,
    SortingValues,
)
from fastapi_filters_standard.utils import fields_include_exclude

TSelectable = TypeVar("TSelectable", bound=Select[Any])


def _overlap(a: Any, b: Any) -> Any:
    try:
        return a.overlap(b)
    except AttributeError:
        return a.overlaps(b)


DEFAULT_FILTERS: Mapping[AbstractFilterOperator, Callable[[Any, Any], Any]] = {
    FilterOperator.eq: operator.eq,
    FilterOperator.ne: operator.ne,
    FilterOperator.gt: operator.gt,
    FilterOperator.ge: operator.ge,
    FilterOperator.lt: operator.lt,
    FilterOperator.le: operator.le,
    FilterOperator.like: lambda a, b: a.like(b),
    FilterOperator.not_like: lambda a, b: ~a.like(b),
    FilterOperator.ilike: lambda a, b: a.ilike(b),
    FilterOperator.not_ilike: lambda a, b: ~a.ilike(b),
    FilterOperator.in_: lambda a, b: a.in_(b),
    FilterOperator.not_in: lambda a, b: a.not_in(b),
    FilterOperator.is_null: lambda a, b: a.is_(None) if b else a.isnot(None),
    FilterOperator.overlap: lambda a, b: _overlap(a, b),
    FilterOperator.not_overlap: lambda a, b: ~_overlap(a, b),
    FilterOperator.contains: lambda a, b: a.contains(b),
    FilterOperator.not_contains: lambda a, b: ~a.contains(b),
}
SORT_FUNCS: Mapping[
    SortingDirection,
    Callable[[ColumnExpressionArgument[Any]], ColumnExpressionArgument[Any]],
] = {
    "asc": asc,
    "desc": desc,
}
SORT_NULLS_FUNCS: Mapping[
    tuple[SortingDirection, SortingNulls],
    Callable[[ColumnExpressionArgument[Any]], ColumnExpressionArgument[Any]],
] = {
    ("asc", "bigger"): nulls_last,
    ("asc", "smaller"): nulls_first,
    ("desc", "bigger"): nulls_first,
    ("desc", "smaller"): nulls_last,
}

EntityNamespace: TypeAlias = Mapping[str, Any]
AdditionalNamespace: TypeAlias = Mapping[str | FilterField[Any], Any]


def _get_entity_namespace(stmt: TSelectable) -> EntityNamespace:
    ns = {}

    for entity in reversed(stmt.get_final_froms()):
        for name, clause in reversed(entity.c.items()):
            ns[name] = clause
            ns[clause.name] = clause

            with suppress(AttributeError):
                table_name = clause.table.name
                ns[f"{table_name}.{clause.name}"] = clause

                if table_name.endswith("s"):
                    ns[f"{table_name[:-1]}.{clause.name}"] = clause

    return ns


def _normalize_additional_namespace(additional: AdditionalNamespace) -> EntityNamespace:
    ns = {}
    for key, value in additional.items():
        k = key.name if isinstance(key, FilterField) else key
        assert k, "Additional namespace key cannot be empty"

        ns[k] = value

    return ns


def _default_hook(*_: Any) -> Any:
    raise NotImplementedError


ApplyFilterFunc: TypeAlias = Callable[
    [TSelectable, EntityNamespace, str, AbstractFilterOperator, Any],
    TSelectable,
]
AddFilterConditionFunc: TypeAlias = Callable[[TSelectable, str, Any], TSelectable]

custom_apply_filter: ConfigVar[ApplyFilterFunc[Any]] = ConfigVar(
    "apply_filter",
    default=_default_hook,
)
custom_add_condition: ConfigVar[AddFilterConditionFunc[Any]] = ConfigVar(
    "add_condition",
    default=_default_hook,
)


def generic_condition(left: Any, right: Any, op: AbstractFilterOperator) -> Any:
    return DEFAULT_FILTERS[op](left, right)


def _apply_filter(
    stmt: TSelectable,
    ns: EntityNamespace,
    field: str,
    op: AbstractFilterOperator,
    val: Any,
    apply_filter: ApplyFilterFunc[TSelectable] | None = None,
    add_condition: AddFilterConditionFunc[TSelectable] | None = None,
) -> TSelectable:
    custom_apply_filter_impl = custom_apply_filter.get()

    try:
        cond = None
        if apply_filter:
            try:
                cond = apply_filter(stmt, ns, field, op, val)
                assert cond is not None
            except (NotImplementedError, AssertionError):
                pass

        if cond is None:
            cond = custom_apply_filter_impl(stmt, ns, field, op, val)
            assert cond is not None
    except (NotImplementedError, AssertionError):
        if field not in ns:
            raise ValueError(f"Unknown field {field}") from None

        try:
            cond = generic_condition(ns[field], val, op)
        except KeyError:
            raise NotImplementedError(f"Operator {op} is not implemented") from None

    if add_condition:
        try:
            return add_condition(stmt, field, cond)
        except NotImplementedError:
            pass

    global_add_condition = custom_add_condition.get()

    try:
        res = global_add_condition(stmt, field, cond)
        assert res is not None

        return cast(TSelectable, res)
    except (NotImplementedError, AssertionError):
        pass

    return stmt.where(cond)  # type: ignore[arg-type]


def apply_filters(
    stmt: TSelectable,
    filters: FilterValues | FilterSet,
    *,
    remapping: Mapping[str, str] | None = None,
    additional: AdditionalNamespace | None = None,
    apply_filter: ApplyFilterFunc[TSelectable] | None = None,
    add_condition: AddFilterConditionFunc[TSelectable] | None = None,
) -> TSelectable:
    if isinstance(filters, FilterSet):
        filters = filters.filter_values

    remapping = remapping or {}
    ns = {
        **_get_entity_namespace(stmt),
        **_normalize_additional_namespace(additional or {}),
    }

    for field, field_filters in filters.items():
        field = remapping.get(field, field)

        for op, val in field_filters.items():
            stmt = _apply_filter(stmt, ns, field, op, val, apply_filter, add_condition)

    return stmt


def apply_sorting(
    stmt: TSelectable,
    sorting: SortingValues,
    *,
    remapping: Mapping[str, str] | None = None,
    additional: AdditionalNamespace | None = None,
) -> TSelectable:
    remapping = remapping or {}
    ns = {
        **_get_entity_namespace(stmt),
        **_normalize_additional_namespace(additional or {}),
    }

    for field, direction, nulls in sorting:
        field = remapping.get(field, field)

        if sort_func := SORT_FUNCS.get(direction):
            sort_expr = sort_func(ns[field])

            if nulls is not None:
                sort_expr = SORT_NULLS_FUNCS[(direction, nulls)](sort_expr)

            stmt = stmt.order_by(sort_expr)
        else:
            raise ValueError(f"Unknown sorting direction {direction}")

    return stmt


def apply_filters_and_sorting(
    stmt: TSelectable,
    filters: FilterValues | FilterSet,
    sorting: SortingValues,
    *,
    remapping: Mapping[str, str] | None = None,
    additional: AdditionalNamespace | None = None,
    apply_filter: ApplyFilterFunc[TSelectable] | None = None,
    add_condition: AddFilterConditionFunc[TSelectable] | None = None,
) -> TSelectable:
    stmt = apply_filters(
        stmt,
        filters,
        remapping=remapping,
        additional=additional,
        apply_filter=apply_filter,
        add_condition=add_condition,
    )
    return apply_sorting(
        stmt,
        sorting,
        remapping=remapping,
        additional=additional,
    )


def adapt_sqlalchemy_column_type(column: ColumnProperty[Any]) -> FilterFieldDef:
    expr: Any = column.expression

    type_: Any
    type_ = list[expr.type.item_type.python_type] if isinstance(expr.type, ARRAY) else expr.type.python_type

    if expr.nullable:
        type_ = type_ | None

    return cast(FilterFieldDef, type_)


def _iter_over_orm_columns(
    obj: Any,
    *,
    include_fk: bool = False,
    include: Container[str] | None = None,
    exclude: Container[str] | None = None,
    remapping: Mapping[str, str] | None = None,
) -> Iterator[tuple[str, ColumnProperty[Any]]]:
    inspected = inspect(obj, raiseerr=True)

    remapping = remapping or {}
    checker = fields_include_exclude(inspected.mapper.attrs.keys(), include, exclude)

    for name, column in inspected.mapper.attrs.items():
        name = remapping.get(name, name)

        if not checker(name):
            continue

        if not isinstance(column, ColumnProperty):
            continue

        if not include_fk and getattr(column.expression, "foreign_keys", None):
            continue

        yield name, column


def create_filters_from_orm(
    obj: Any,
    *,
    in_: FilterPlace | None = None,
    alias_generator: FilterAliasGenerator | None = None,
    include_fk: bool = False,
    include: Container[str] | None = None,
    exclude: Container[str] | None = None,
    remapping: Mapping[str, str] | None = None,
    **overrides: FilterFieldDef,
) -> FiltersResolver:
    fields = {
        name: adapt_sqlalchemy_column_type(column)
        for name, column in _iter_over_orm_columns(
            obj,
            include_fk=include_fk,
            include=include,
            exclude=exclude,
            remapping=remapping,
        )
    }

    return create_filters(
        in_=in_,
        alias_generator=alias_generator,
        **{**fields, **overrides},
    )


def create_sorting_from_orm(
    obj: Any,
    *,
    default: str | None = None,
    in_: FilterPlace | None = None,
    include_fk: bool = False,
    include: Container[str] | None = None,
    exclude: Container[str] | None = None,
    remapping: Mapping[str, str] | None = None,
) -> SortingResolver:
    fields = [
        name
        for name, _ in _iter_over_orm_columns(
            obj,
            include_fk=include_fk,
            include=include,
            exclude=exclude,
            remapping=remapping,
        )
    ]

    return create_sorting(
        *fields,
        in_=in_,
        default=default,
    )


__all__ = [
    "adapt_sqlalchemy_column_type",
    "apply_filters",
    "apply_filters_and_sorting",
    "apply_sorting",
    "create_filters_from_orm",
    "create_sorting_from_orm",
    "custom_add_condition",
    "custom_apply_filter",
    "generic_condition",
]



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    overload,
)

from .op import FilterOpBuilder
from .operators import FilterOperator, get_filter_operators
from .utils import is_seq

if TYPE_CHECKING:
    from .types import AbstractFilterOperator


T_co = TypeVar("T_co", covariant=True)


@dataclass(eq=False, order=False)
class FilterField(FilterOpBuilder[T_co]):
    type: type[T_co] | None = None
    operators: list[AbstractFilterOperator] | None = None
    default_op: AbstractFilterOperator | None = None
    name: str | None = None
    alias: str | None = None
    internal: bool = False
    op_types: dict[AbstractFilterOperator, Any] | None = None

    if TYPE_CHECKING:

        @overload
        def __get__(self, instance: None, owner: Any) -> FilterField[T_co]:
            pass

        @overload
        def __get__(
            self,
            instance: object,
            owner: Any,
        ) -> dict[AbstractFilterOperator, T_co | Sequence[T_co] | bool]:
            pass

        def __get__(self, instance: object | None, owner: Any) -> Any:
            pass

    def __hash__(self) -> int:
        return hash((self.name, self.alias, self.type))

    def __set_name__(self, owner: Any, name: str) -> None:
        self.name = name

    def __post_init__(self) -> None:
        if self.operators is None and self.type is not None:
            self.operators = [*get_filter_operators(self.type)]

        if self.default_op is None:
            if is_seq(self.type):
                self.default_op = FilterOperator.overlap
            else:
                self.default_op = FilterOperator.eq


__all__ = [
    "FilterField",
]



from collections.abc import Mapping
from typing import Any, TypeVar, cast

from tortoise.queryset import QuerySet

from fastapi_filters_standard import FilterOperator, FilterSet, FilterValues
from fastapi_filters_standard.types import AbstractFilterOperator, SortingValues

TStmt = TypeVar("TStmt", bound=QuerySet[Any])


DEFAULT_FILTERS: Mapping[AbstractFilterOperator, str] = {
    FilterOperator.eq: "",
    FilterOperator.ne: "__not",
    FilterOperator.gt: "__gt",
    FilterOperator.ge: "__ge",
    FilterOperator.lt: "__lt",
    FilterOperator.le: "__le",
    FilterOperator.like: "__like",
    FilterOperator.not_like: "__not_like",
    FilterOperator.ilike: "__ilike",
    FilterOperator.not_ilike: "__not_ilike",
    FilterOperator.in_: "__in",
    FilterOperator.not_in: "__not_in",
    FilterOperator.is_null: "__is_null",
    FilterOperator.contains: "__contains",
    FilterOperator.not_contains: "__not_contains",
}


def apply_filters(
    stmt: TStmt,
    filters: FilterValues | FilterSet,
    *,
    remapping: Mapping[str, str] | None = None,
) -> TStmt:
    remapping = remapping or {}
    if isinstance(filters, FilterSet):
        filters = filters.filter_values

    for field, field_filters in filters.items():
        field = remapping.get(field, field)
        field = field.replace(".", "__")

        for op, val in field_filters.items():
            if (cond := DEFAULT_FILTERS.get(op)) is not None:
                stmt = cast(TStmt, stmt.filter(**{f"{field}{cond}": val}))
            else:
                raise NotImplementedError(f"Operator {op} is not implemented")

    return stmt


def apply_sorting(
    stmt: TStmt,
    sorting: SortingValues,
    remapping: Mapping[str, str] | None = None,
) -> TStmt:
    remapping = remapping or {}
    ordering = []
    for field, direction, _ in sorting:
        field = remapping.get(field, field)
        field = field.replace(".", "__")
        ordering.append(f"{'-' if direction == 'desc' else ''}{field}")

    if ordering:
        stmt = stmt.order_by(*ordering)  # type: ignore[assignment]

    return stmt


def apply_filters_and_sorting(
    stmt: TStmt,
    filters: FilterValues | FilterSet,
    sorting: SortingValues,
    *,
    remapping: Mapping[str, str] | None = None,
) -> TStmt:
    stmt = apply_filters(stmt, filters, remapping=remapping)
    return apply_sorting(stmt, sorting, remapping=remapping)


__all__ = [
    "apply_filters",
    "apply_filters_and_sorting",
    "apply_sorting",
]
